import numpy as np
import re
import sys
import os
import datetime
import time
import calendar
import verif.input
import matplotlib.dates
import matplotlib.ticker
import verif.aggregator
import verif.field
import verif.util
import verif.variable
import scipy


class Data(object):
    """ Organizes data from several inputs

    Access verification data from a list of verif.input. Only returns data that
    is available for all files, for fair comparisons i.e if some
    times/leadtimes/locations are missing.

    Instance attribute:
    times          A numpy array of available initialization times
    leadtimes      A numpy array of available leadtimes
    locations      A list of available locations
    thresholds     A numpy array of available thresholds
    quantiles      A numpy array of available quantiles
    num_inputs     The number of inputs in the dataset
    variable       The variable
    timesofday     Available times of day (derived from times)
    daysofyear     Available days of year (derived from times)
    days           Available days (derived from times)
    weeks          Available weeks (derived from times)
    months         Available months (derived from times)
    years          Available years (derived from times)
    leadtimedays   Available leadtimedays (derived from leadtimes)
    """
    def __init__(self, inputs, times=None, dates=None, tods=None, leadtimes=None, locations=None, locations_x=None,
          lat_range=None, lon_range=None, elev_range=None, clim=None, clim_type="subtract",
          obs_range=None, legend=None, remove_missing_across_all=True,
          obs_field=verif.field.Obs(),
          fcst_field=verif.field.Fcst(),
          dim_agg_length=None,
          dim_agg_axis=verif.axis.Leadtime(),
          dim_agg_method=verif.aggregator.Mean()):

        """
        Arguments:
        inputs         A list of verif.input
        times          A numpy array of times. Discard data for all other times
        dates          A numpy array of dates. Only allow times that occur on
                       these dates (but alllow any hour of the day).
        tods           A numpy array of time of day values (in hours). Only allow
                       initialization times for these hours of the day (in UTC).
        leadtimes      A numpy array of leadtimes. Discard data for all other leadtimes
        locations      A list of verif.location. Discard data for all other locations
        locations_x    A list of verif.location to not remove
        clim           Use this verif.input to compute anomaly. Should therefore
                       be a climatological forecast. Subtract/divide the
                       forecasts from this file from all forecasts and
                       observations from the other files.
        clim_type      Operation to apply with climatology. Either 'subtract', or
                       'divide'
        dim_agg_length Aggregate obs and fcst across this many leadtimes
        """

        if not isinstance(inputs, list):
            inputs = [inputs]
        self._remove_missing_across_all = remove_missing_across_all

        if legend is not None and len(inputs) is not len(legend):
            verif.util.error("Need one legend entry for each filename")
        self._legend = legend

        self._obs_field = obs_field
        self._fcst_field = fcst_field

        self._obs_range = obs_range

        # Organize inputs
        self._inputs = list()
        self._get_score_cache = list()  # Caches data from input
        self._get_scores_cache = dict()  # Caches the output from get_scores
        self._clim = None
        for input in inputs:
            self._inputs.append(input)
            self._get_score_cache.append(dict())
        if clim is not None:
            self._clim = clim
            self._get_score_cache.append(dict())
            if not (clim_type == "subtract" or clim_type == "divide"):
                verif.util.error("Data: clim_type must be 'subtract' or 'divide")
            self._clim_type = clim_type

            # Add climatology to the end
            self._inputs = self._inputs + [self._clim]

        # Latitude-Longitude range
        if lat_range is not None or lon_range is not None:
            lat = [loc.lat for loc in self._inputs[0].locations]
            lon = [loc.lon for loc in self._inputs[0].locations]
            loc_id = [loc.id for loc in self._inputs[0].locations]
            latlon_locations = list()
            min_lon = -180
            max_lon = 180
            min_lat = -90
            max_lat = 90
            if lat_range is not None:
                min_lat = lat_range[0]
                max_lat = lat_range[1]
            if lon_range is not None:
                min_lon = lon_range[0]
                max_lon = lon_range[1]
            for i in range(0, len(lat)):
                currLat = float(lat[i])
                currLon = float(lon[i])
                if currLat >= min_lat and currLat <= max_lat and currLon >= min_lon and currLon <= max_lon:
                    latlon_locations.append(loc_id[i])
            use_locations = list()
            if locations is not None:
                use_locations = [loc for loc in locations if loc in latlon_locations]
            else:
                use_locations = latlon_locations
            if len(use_locations) == 0:
                verif.util.error("No available locations within lat/lon range")
        elif locations is not None:
            use_locations = locations
        else:
            use_locations = [s.id for s in self._inputs[0].locations]

        # Elevation range
        if elev_range is not None:
            locations = self._inputs[0].locations
            min_elev = elev_range[0]
            max_elev = elev_range[1]
            elev_locations = list()
            for i in range(0, len(locations)):
                curr_elev = float(locations[i].elev)
                id = locations[i].id
                if curr_elev >= min_elev and curr_elev <= max_elev:
                    elev_locations.append(id)
            use_locations = verif.util.intersect(use_locations, elev_locations)
            if len(use_locations) == 0:
                verif.util.error("No available locations within elevation range")

        # Remove locations
        if locations_x is not None:
            use_locations = [loc for loc in use_locations if loc not in locations_x]

        # Find common indicies
        # TODO: Merge in information from dates (but these span whole days)
        self._timesI = self._get_common_indices(self._inputs, verif.axis.Time(), times)
        self._leadtimesI = self._get_common_indices(self._inputs, verif.axis.Leadtime(), leadtimes)
        self._locationsI = self._get_common_indices(self._inputs, verif.axis.Location(), use_locations)
        if len(self._timesI[0]) == 0:
            verif.util.error("No valid times selected")
        if len(self._leadtimesI[0]) == 0:
            verif.util.error("No valid leadtimes selected")
        if len(self._locationsI[0]) == 0:
            verif.util.error("No valid locations selected")

        # Load dimension information
        self.times = self._get_times()
        self.leadtimes = self._get_leadtimes()
        self.locations = self._get_locations()
        self.thresholds = self._get_thresholds()
        self.quantiles = self._get_quantiles()
        self.variable = self._get_variable()
        self.num_inputs = self._get_num_inputs()

        if dates is not None:
            """
            Only allow specified dates. Allow all initialization times that are
            within each of the dates.
            """
            dates_times = [verif.util.date_to_unixtime(t) for t in dates]
            self.times = np.array([t for t in self.times if int(t / 86400)*86400 in dates_times])

        if tods is not None:
            self.times = np.array([t for t in self.times if int(t % 86400)/3600 in tods])
        """
        Indices must be recomputed since some times are removed by -d and -tod.
        There doesn't seem to be an easy way to avoid this.
        """
        self._timesI = self._get_common_indices(self._inputs, verif.axis.Time(), self.times)

        # Compute axis values
        self.axis_cache = dict()
        self.axis_cache_unique = dict()
        for axis in verif.axis.get_time_axes():
            self.axis_cache[axis] = axis.compute_from_times(self.times)
            self.axis_cache_unique[axis] = np.unique(self.axis_cache[axis])
        for axis in verif.axis.get_leadtime_axes():
            self.axis_cache[axis] = axis.compute_from_leadtimes(self.leadtimes)
            self.axis_cache_unique[axis] = np.unique(self.axis_cache[axis])

        self.dim_agg_length = dim_agg_length
        self.dim_agg_axis = dim_agg_axis
        self.dim_agg_method = dim_agg_method

    def get_fields(self):
        """ Get a list of fields that all inputs have

        Returns:
           list(verif.field.Field): A list of fields
        """
        all_fields = set()
        for f in range(self._get_num_inputs_with_clim()):
            input = self._inputs[f]
            if f == 0:
                all_fields = input.get_fields()
            else:
                all_fields = set(all_fields) & set(input.get_fields())
        return list(all_fields)

    def get_scores(self, fields, input_index, axis=verif.axis.All(), axis_index=None):
        """ Retrieves scores from all files

        Climatology is handled by subtracting clim's fcst field from any
        obs or determinsitic fields.

        Arguments:
        fields         Which verif.field should be retreived? Either a list or a
                       single field can be supplied, however the return type will
                       match this choice.
        input_index    Which input to pull from? Must be between 0 and num_inputs
        axis           Which axis to aggregate against. If verif.axis.All() is
                       used, then no aggregation takes place and the 3D numpy
                       array is returned.
        axis_index     Which slice along the axis to retrieve

        Returns:
        scores         A list of numpy arrays. If fields is not a list, but a
                       single field, then a numpy array is returned.
        """

        fields_is_single = False
        if not isinstance(fields, list):
            fields = [fields]
            fields_is_single = True

        key = (tuple(fields), input_index, axis, axis_index)
        if key in self._get_scores_cache.keys():
            if fields_is_single:
                return self._get_scores_cache[key][0]
            else:
                return self._get_scores_cache[key]

        if input_index < 0 or input_index >= self.num_inputs:
            verif.util.error("input_index must be between 0 and %d" % self.num_inputs)

        scores = list()
        valid = None

        # Compute climatology, if needed
        obsFcstAvailable = (verif.field.Obs() in fields or verif.field.Fcst() in fields)
        doClim = self._clim is not None and obsFcstAvailable
        if doClim:
            temp = self._get_score(verif.field.Fcst(), len(self._inputs) - 1)
            clim = self._apply_axis(temp, axis, axis_index)
        else:
            clim = 0

        # Load scores and flatten along the correct dimension
        for i in range(0, len(fields)):
            field = fields[i]
            temp = self._get_score(field, input_index)

            # Remove observations outside the obsrange
            if self._obs_range is not None and field == verif.field.Obs():
                with np.errstate(invalid='ignore'):
                    temp[temp < self._obs_range[0]] = np.nan
                    temp[temp > self._obs_range[1]] = np.nan

            curr = self._apply_axis(temp, axis, axis_index)

            # Subtract climatology
            if doClim and (field == verif.field.Fcst() or field == verif.field.Obs()):
                if self._clim_type == "subtract":
                    curr = curr - clim
                else:
                    curr = curr / clim

            # Remove missing values
            currValid = (np.isnan(curr) == 0) & (np.isinf(curr) == 0)

            # Collapse the ensemble dimension, by checking that all members are available
            if field == verif.field.Ensemble():
                num_members = currValid.shape[-1]
                if num_members == 0:
                    currValid = np.ones(currValid.shape[:-1], int)
                else:
                    if axis == verif.axis.All():
                        currValid = np.prod(currValid, axis=3)
                    else:
                        currValid = np.prod(currValid, axis=1)

            if valid is None:
                valid = currValid
            else:
                valid = (valid & currValid)
            scores.append(curr)

        if axis == verif.axis.All():
            I = np.where(valid == 0)
            for i in range(0, len(fields)):
                if len(I) == 3:
                    scores[i][I[0], I[1], I[2]] = np.nan
                elif len(I) == 4:
                    # Ensemble forecasts
                    scores[i][I[0], I[1], I[2], I[3]] = np.nan
                else:
                    raise Exception("Internal error")
        else:
            I = np.where(valid)[0]
            for i in range(0, len(fields)):
                # Use ... as this might be an ensemble field
                scores[i] = scores[i][I, ...]

        # No valid data. Therefore return a list of nans instead of an empty list
        if scores[0].shape[0] == 0:
            new_scores = list()
            for i, field in enumerate(fields):
                if field == verif.field.Ensemble():
                    num_members = scores[i].shape[-1]
                    new_scores = [np.nan * np.zeros([1, num_members], float)]
                else:
                    new_scores += [np.nan * np.zeros(1, float)]
            scores = new_scores

        self._get_scores_cache[key] = scores

        # Turn into a single numpy array if we were not supplied with a list of
        # fields
        if fields_is_single:
            return scores[0]
        else:
            return scores

    def get_axis_size(self, axis):
        return len(self.get_axis_values(axis))

    def get_axis_values(self, axis):
        """ What are the values along an axis?
        verif.axis.Time()          Unixtimes
        verif.axis.Month()         Unixtimes of the begining of each month
        verif.axis.Year()          Unixtimes of the beginning of each year
        verif.axis.Leadtime()      Lead times in hours
        verif.axis.Leadtimeday()   Lead time day in days
        verif.axis.Timeofday()     Time of day in hours
        verif.axis.Location()      Location id
        verif.axis.Lat()           Latitudes of locations
        verif.axis.Lon()           Longitudes of locations
        verif.axis.Elev()          Elevations of locations

        Arguments:
        axis        of type verif.axis.Axis

        Returns:
        array       a 1D numpy array of values
        """
        if axis == verif.axis.Time():
            return self.times
        elif axis in verif.axis.get_time_axes():
            return np.unique(axis.compute_from_times(self.times))
        elif axis in verif.axis.get_leadtime_axes():
            return np.unique(axis.compute_from_leadtimes(self.leadtimes))
        elif axis == verif.axis.No():
            return [0]
        elif axis.is_location_like:
            if axis == verif.axis.Location():
                data = np.array([loc.id for loc in self.locations])
            elif axis == verif.axis.Elev():
                data = np.array([loc.elev for loc in self.locations])
            elif axis == verif.axis.Lat():
                data = np.array([loc.lat for loc in self.locations])
            elif axis == verif.axis.Lon():
                data = np.array([loc.lon for loc in self.locations])
            else:
                verif.util.error("Data.get_axis_values has a bad axis name: " + axis)
            return data
        else:
            return [0]

    def get_axis_locator(self, axis):
        """ Where should ticks be located for this axis? Returns an mpl Locator """
        if axis == verif.axis.Leadtime():
            # Define our own locators, since in general we want multiples of 24
            # (or even fractions thereof) to make the ticks repeat each day. Aim
            # for a maximum of 12 ticks.
            leadtimes = self.get_axis_values(verif.axis.Leadtime())
            span = max(leadtimes) - min(leadtimes)
            if span > 300:
                return matplotlib.ticker.AutoLocator()
            elif span > 200:
                return matplotlib.ticker.MultipleLocator(48)
            elif span > 144:
                return matplotlib.ticker.MultipleLocator(24)
            elif span > 72:
                return matplotlib.ticker.MultipleLocator(12)
            elif span > 36:
                return matplotlib.ticker.MultipleLocator(6)
            elif span > 12:
                return matplotlib.ticker.MultipleLocator(3)
            else:
                return matplotlib.ticker.MultipleLocator(1)
        else:
            return matplotlib.ticker.AutoLocator()

    def get_full_names(self):
        names = [input.fullname for input in self._inputs]
        if self._clim is not None:
            names = names[0:-1]
        return names

    def get_names(self):
        names = [input.name for input in self._inputs]
        if self._clim is not None:
            names = names[0:-1]
        return names

    def get_short_names(self):
        names = [input.shortname for input in self._inputs]
        if self._clim is not None:
            names = names[0:-1]
        return names

    def get_num_members(self, f):
        return self._inputs[f].num_members

    def get_legend(self):
        if self._legend is None:
            legend = self.get_short_names()
        else:
            legend = self._legend
        return legend

    def get_variable_and_units(self):
        var = self.variable
        return var.name + " (" + var.units + ")"

    def get_axis_descriptions(self, axis):
        if axis.is_location_like:
            q = dict()
            descs = list()
            ids = [loc.id for loc in self.locations]
            lats = [loc.lat for loc in self.locations]
            lons = [loc.lon for loc in self.locations]
            elevs = [loc.elev for loc in self.locations]
            return {"id": ids, "lat": lats, "lon": lons, "elev": elevs}
        elif axis.is_time_like:
            unixtimes = self.get_axis_values(axis)
            # Convert to date objects
            dates = [matplotlib.dates.num2date(verif.util.unixtime_to_datenum(unixtime)) for unixtime in unixtimes]
            times = list()
            fmt = axis.fmt

            times = [date.strftime(fmt) for date in dates]
            return {axis.name(): times}
        else:
            return {axis.name(): self.get_axis_values(axis)}

    def _get_score(self, field, input_index):
        """ Load the field variable from input, but only include the common data

        Scores loaded will have the same dimension, regardless what input_index
        is used.

        field:         The type is of verif.field
        input_index:   which input to load from
        """

        num_inputs = self._get_num_inputs_with_clim()
        if field == verif.field.Obs():
            """
            Treat observation different. Since the observations should be the same
            for all inputs, reuse the observations if one or more inputs are
            missing these
            """
            field = self._obs_field
            if field in self._get_score_cache[input_index]:
                return self._get_score_cache[input_index][field]

            found_obs = False
            for i in range(num_inputs):
                input = self._inputs[i]
                if input.has_field(field):
                    if field == verif.field.Obs():
                        temp = input.obs
                        temp = self.preaggregate(temp, input)
                    elif field == verif.field.Fcst():
                        temp = input.fcst
                        temp = self.preaggregate(temp, input)
                    else:
                        temp = input.other_score(field.name())
                        if self.dim_agg_length is not None:
                            verif.util.warning("Cannot preaggregate " + field.name() + "since I'm unsure it makes sense")

                    Itimes = self._get_time_indices(i)
                    Ileadtimes = self._get_leadtime_indices(i)
                    Ilocations = self._get_location_indices(i)
                    temp = temp[Itimes, :, :]
                    temp = temp[:, Ileadtimes, :]
                    temp = temp[:, :, Ilocations]

                    self._get_score_cache[i][field] = temp
                    found_obs = True
            if not found_obs:
                verif.util.error("No files have observations")

            for i in range(num_inputs):
                if field not in self._get_score_cache[i]:
                    for j in range(num_inputs):
                        if field in self._get_score_cache[j]:
                            verif.util.warning("No observations in %s. Loading from %s" % (self._inputs[i].fullname, self._inputs[j].fullname))
                            self._get_score_cache[i][field] = self._get_score_cache[j][field]
                            break
        else:
            # Check if data is cached
            if field in self._get_score_cache[input_index]:
                return self._get_score_cache[input_index][field]

            if field == verif.field.Fcst():
                field = self._fcst_field

            for i in range(num_inputs):
                if field not in self._get_score_cache[i]:
                    input = self._inputs[i]
                    if isinstance(field, verif.field.Threshold):
                        I = np.where(np.isclose(input.thresholds,  field.threshold))[0]
                        if len(I) == 0 or self.dim_agg_length is not None:
                            # Try to get probabilities from ensemble
                            if input.ensemble is None:
                                verif.util.error("%s does not contain '%s'" % (self.get_names()[i], field.name()))
                            temp = self.preaggregate(input.ensemble, input)

                            # TODO: Use interpolation here, which is better for small ensemble
                            # sizes, and would be more consistent with how this is done with
                            # computing quantiles. The problem is that there is no function
                            # available in scipy or numpy for interpolating where x is
                            # multi-dimensional and y is a vector.

                            # NOTE: <= operator does not handle nans
                            is_missing = np.isnan(temp)
                            temp = (temp <= field.threshold).astype(np.float32)
                            temp[is_missing] = np.nan
                            temp = np.nanmean(temp, axis=-1)
                        else:
                            if self.dim_agg_length is not None:
                                verif.util.error("Cannot pre-aggregate threshold probabilities without an ensemble")
                            assert(len(I) == 1)
                            temp = input.threshold_scores[:, :, :, I[0]]

                    elif isinstance(field, verif.field.Quantile):
                        I = np.where(np.isclose(input.quantiles, field.quantile))[0]
                        if len(I) == 0 or self.dim_agg_length is not None:
                            if input.ensemble is None:
                                verif.util.error("%s does not contain '%s'" % (self.get_names()[i], field.name()))
                            num_members = input.ensemble.shape[-1]
                            if field.quantile < get_lower_cdf(num_members) or field.quantile > get_upper_cdf(num_members):
                                verif.util.warning("In %s, ensemble doesn't have enough members to accurately get quantile level %s" % (self.get_names()[i], field.quantile))

                            temp = self.preaggregate(input.ensemble, input)
                            temp = np.quantile(input.ensemble, field.quantile, axis=3, method="normal_unbiased")
                        else:
                            if self.dim_agg_length is not None:
                                verif.util.error("Cannot pre-aggregate threshold probabilities without an ensemble")
                            assert(len(I) == 1)
                            temp = input.quantile_scores[:, :, :, I[0]]

                    elif field == verif.field.EnsembleVariance():
                        if self.dim_agg_length is not None:
                            temp = self.preaggregate(input.ensemble, input)
                            temp = np.nanvar(temp, axis=-1)
                        else:
                            temp = input.ensemble_variance

                    elif field == verif.field.Pit():
                        temp = input.pit
                        x0 = self.variable.x0
                        x1 = self.variable.x1
                        if x0 is not None or x1 is not None:
                            temp = verif.field.Pit.randomize(input.obs, temp, x0, x1)

                    elif input.has_field(field):
                        # Fields that can be preaggregated

                        if field == verif.field.Obs():
                            temp = input.obs
                            temp = self.preaggregate(temp, input)

                        elif field == verif.field.Fcst():
                            temp = input.fcst
                            temp = self.preaggregate(temp, input)

                        elif field == verif.field.EnsembleMean():
                            temp = input.ensemble_mean
                            temp = self.preaggregate(temp, input)

                        elif isinstance(field, verif.field.EnsembleMember):
                            if field.member >= input.num_members:
                                temp = np.nan * np.zeros(input.ensemble[:, :, :, 0].shape)
                            else:
                                temp = input.ensemble[:, :, :, field.member]
                            temp = self.preaggregate(temp, input)

                        elif isinstance(field, verif.field.Ensemble):
                            temp = input.ensemble[:, :, :, :]
                            temp = self.preaggregate(temp, input)

                        else:
                            temp = input.other_score(field.name())
                            if self.dim_agg_length is not None:
                                verif.util.warning("Cannot preaggregate " + field.name() + "since I'm unsure it makes sense")
                    elif isinstance(field, verif.field.Ensemble):
                        # An input can potentially not have an ensemble, but we still want to
                        # process the empy ensemble
                        temp = input.ensemble
                    else:
                        verif.util.error("%s does not contain '%s'" % (self.get_names()[i], field.name()))

                    Itimes = self._get_time_indices(i)
                    Ileadtimes = self._get_leadtime_indices(i)
                    Ilocations = self._get_location_indices(i)
                    temp = temp[Itimes, ...]
                    temp = temp[:, Ileadtimes, ...]
                    temp = temp[:, :, Ilocations, ...]

                    self._get_score_cache[i][field] = temp

        """
        Remove missing. If one configuration has a missing value, set all
        configurations to missing. This can happen when the times are
        available, but have missing values.

        Ensembles can have different number of members and some members can be missing on some days.
        For this field, only remove cases where one or more input files are missing all members.
        """
        if self._remove_missing_across_all:
            def compute_is_missing(array):
                if array.shape[-1] == 0:
                    # When an ensemble has no members, it's technically never missing
                    return np.zeros(array.shape[0:-1]) > 1
                else:
                    return np.all(np.isnan(array), axis=3)

            if isinstance(field, verif.field.Ensemble):
                curr_values = self._get_score_cache[0][field]
                num_members = curr_values.shape[3]
                # Check if all members are missing (axis 3)
                is_missing = compute_is_missing(curr_values)
                for i in range(1, num_inputs):
                    is_missing = is_missing | (compute_is_missing(self._get_score_cache[i][field]))
                for i in range(num_inputs):
                    temp = self._get_score_cache[i][field]
                    # Remove all members by broadcasting the True/False to all members
                    I = np.repeat(is_missing[..., None], temp.shape[-1], 3)
                    self._get_score_cache[i][field][I] = np.nan
            else:
                is_missing = np.isnan(self._get_score_cache[0][field])
                for i in range(1, num_inputs):
                    is_missing = is_missing | (np.isnan(self._get_score_cache[i][field]))
                for i in range(num_inputs):
                    self._get_score_cache[i][field][is_missing] = np.nan

        return self._get_score_cache[input_index][field]

    def _calculate_window(self, array, leadtimes):
        O = array.shape[1]
        Inan = np.isnan(array)
        for o in range(0, O):
            threshold = 0.5
            q = np.nansum(np.cumsum(array[:, o:, :], axis=1) <= threshold, axis=1)
            I = q+o
            I[I >= O] = O-1
            array[:, o, :] = leadtimes[I] - leadtimes[o]
        array[Inan] = np.nan
        # array[array > 2] = 2
        return array

    def _get_times(self):
        times = self._inputs[0].times
        I = self._timesI[0]
        return np.array([times[i] for i in I], int)

    def _get_leadtimes(self):
        leadtimes = self._inputs[0].leadtimes
        I = self._leadtimesI[0]
        return np.array([leadtimes[i] for i in I], float)

    def _get_locations(self):
        locations = self._inputs[0].locations
        I = self._locationsI[0]
        use_locations = list()
        for i in I:
            use_locations.append(locations[i])
        return use_locations

    @staticmethod
    def _get_common_indices(inputs, axis, aux=None):
        """
        Find indicies of elements that are present in all inputs. Merge in values
        in 'aux' as well

        Arguments:
        inputs      A list of verif.input
        axis        An axis of type verif.axis

        Returns a list of arrays, one array for each input
        """
        # Find common values among all inputs
        available_values = aux
        for input in inputs:
            if axis == verif.axis.Time():
                curr_values = input.times
            elif axis == verif.axis.Leadtime():
                curr_values = input.leadtimes
            elif axis == verif.axis.Location():
                locations = input.locations
                curr_values = [loc.id for loc in locations]

            curr_values = np.sort(curr_values)
            len_before = len(curr_values)
            curr_values = np.unique(curr_values)

            if len(curr_values) != len_before:
                verif.util.warning("The '%s' axis in '%s' has repeated values.  Using the first value." % (axis.name(), input.fullname))

            if available_values is None:
                available_values = curr_values
            else:
                # if axis == verif.axis.Leadtime():
                #     available_values = np.intersect1d(available_values, curr_values * 1000 // 1 / 1000)
                # else:
                available_values = np.intersect1d(available_values, curr_values)

        # Sort values, since for example, times may not be in an ascending order
        available_values = np.sort(available_values)

        # Remove nan values
        available_values = available_values[np.isnan(available_values) == 0]

        # Determine which index each value is at
        indices = list()
        for input in inputs:
            if axis == verif.axis.Time():
                temp = input.times
            elif axis == verif.axis.Leadtime():
                temp = input.leadtimes
            elif axis == verif.axis.Location():
                locations = input.locations
                temp = np.zeros(len(locations))
                for i in range(0, len(locations)):
                    temp[i] = locations[i].id
            I = np.where(np.isin(temp, available_values))[0]
            II = np.zeros(len(available_values), 'int')
            for i in range(0, len(available_values)):
                III = np.where(available_values[i] == temp)[0]
                assert(len(III) >= 1)
                II[i] = III[0]

            indices.append(II)
        return indices

    def _get_thresholds(self):
        thresholds = None
        for input in self._inputs:
            currThresholds = input.thresholds
            if thresholds is None:
                thresholds = currThresholds
            else:
                thresholds = set(thresholds) & set(currThresholds)

        thresholds = sorted(thresholds)
        thresholds = np.array(thresholds)
        return thresholds

    def _get_quantiles(self):
        quantiles = None
        for input in self._inputs:
            currQuantiles = input.quantiles
            if quantiles is None:
                quantiles = currQuantiles
            else:
                quantiles = set(quantiles) & set(currQuantiles)

        quantiles = sorted(quantiles)
        quantiles = np.array(quantiles)
        return quantiles

    def _get_time_indices(self, input_index):
        return self._timesI[input_index]

    def _get_leadtime_indices(self, input_index):
        return self._leadtimesI[input_index]

    def _get_location_indices(self, input_index):
        return self._locationsI[input_index]

    def _get_num_inputs(self):
        return len(self._inputs) - (self._clim is not None)

    def _get_num_inputs_with_clim(self):
        return len(self._inputs)

    def _get_variable(self):
        # TODO: Only check first file?
        variable = self._inputs[0].variable

        # Handle obs field
        units = self._obs_field.units(variable)
        name = self._obs_field.label(variable)
        formatter = self._obs_field.formatter(variable)
        x0 = variable.x0
        x1 = variable.x1
        variable = verif.variable.Variable(name, units, formatter, x0, x1)

        return variable

    def _apply_axis(self, array, axis, axis_index):
        """ Slice an array along a certain axis and return an extract array

        Arguments:
        array       3D/4D numpy array
        axis        Of type verif.axis.Axis
        axis_index  Index along the axis to slice
        """
        output = None
        if axis == verif.axis.All() or axis is None:
            output = array
        else:
            is_ensemble = len(array.shape) == 4
            if axis == verif.axis.Time():
                output = array[axis_index, ...]
            elif axis in verif.axis.get_time_axes():
                axis_values = self.axis_cache[axis]
                axis_values_unique = self.axis_cache_unique[axis]
                I = np.where(axis_values == axis_values_unique[axis_index])
                output = array[I, ...]
            elif axis in verif.axis.get_leadtime_axes():
                axis_values = self.axis_cache[axis]
                axis_values_unique = self.axis_cache_unique[axis]
                I = np.where(axis_values == axis_values_unique[axis_index])
                output = array[:, I, ...]
            elif axis.is_location_like:
                output = array[:, :, axis_index, ...]
            elif axis in [verif.axis.No(), verif.axis.Threshold(), verif.axis.Obs(), verif.axis.Fcst()]:
                output = array
            else:
                verif.util.error("data.py: unrecognized axis: " + axis.name())

            if is_ensemble:
                # Put into (N, members)
                num_members = array.shape[-1]
                if num_members == 0:
                    N = np.prod(output.shape[:-1])
                    output = np.zeros([N, num_members], 'float')
                else:
                    output = np.array([output[..., i].flatten() for i in range(num_members)])
                    output = output.transpose()
            else:
                output = output.flatten()

        return output

    def preaggregate(self, array, input):
        temp = array
        if self.dim_agg_length is not None:
            if self.dim_agg_axis == verif.axis.Time():
                temp = preaggregate_time(temp, input.times, self.dim_agg_method, self.dim_agg_length)
            elif self.dim_agg_axis == verif.axis.Leadtime():
                temp = preaggregate_leadtime(temp, input.leadtimes, self.dim_agg_method, self.dim_agg_length)
            else:
                verif.util.error("Dimension aggregation has to be one of 'time' or 'leadtime'")
        return temp


def preaggregate_time(array, times, aggregator, length):
    """ Aggregate array across time

    Args:
        array (np.array): 3D array of observations or forecasts
        times (np.array): 1D array of unixtimes (seconds)
        aggregator (verif.aggregator.Aggregator): Aggregate with this function
        length (float): Aggregate across this many hours in time
    """
    assert len(times) == array.shape[0]
    new_array = np.nan * np.zeros(array.shape, np.float32)
    for t in range(array.shape[0]):
        start = times[t] - length * 3600
        I = range(np.where(times > start)[0][0], t+1)
        new_array[t, ...] = aggregator(array[I, ...], axis=0)
    return new_array


def preaggregate_leadtime(array, leadtimes, aggregator, length):
    """ Aggregate array across leadtime

    Args:
        array (np.array): 3D array of observations or forecasts
        leadtimes (np.array): 1D array of leadtimes (hours)
        aggregator (verif.aggregator.Aggregator): Aggregate with this function
        length (float): Aggregate across this many hours in leadtime
    """
    assert len(leadtimes) == array.shape[1]
    new_array = np.nan * np.zeros(array.shape, np.float32)
    for t in range(array.shape[1]):
        start = leadtimes[t] - length
        I = range(np.where(leadtimes > start)[0][0], t+1)
        new_array[:, t, ...] = aggregator(array[:, I, ...], axis=1)
    return new_array


def get_lower_cdf(num_members):
    """Returns the CDF corresponding to the lowest ensemble member"""
    return 0
    # return 1 / (num_members + 1)
    # return 0.5 / num_members


def get_upper_cdf(num_members):
    return 1 - get_lower_cdf(num_members)


def get_quantile_function(ensemble):
    assert len(ensemble.shape) == 4

    num_members = ensemble.shape[3]
    lower_cdf = get_lower_cdf(num_members)
    upper_cdf = get_upper_cdf(num_members)
    ensemble = np.sort(ensemble, axis=-1)

    x = np.linspace(lower_cdf, upper_cdf, num_members)

    f = scipy.interpolate.interp1d(x,
            ensemble, bounds_error=True,
            axis=3, kind='zero')
    return f



import unittest
import numpy as np
import verif.input
import verif.location
import verif.data


class InputNetcdfTest(unittest.TestCase):
    def test_valid2(self):
        input = verif.input.Netcdf("verif/tests/files/netcdf_valid2.nc")
        locations = input.locations
        self.assertTrue(1, len(locations))
        self.assertEqual(
            verif.location.Location(18700, 59.9423, 10.72, 94), locations[0]
        )
        np.testing.assert_array_equal(np.array([0, 1, 2]), input.leadtimes)
        np.testing.assert_array_equal(np.array([1388534400, 1388620800]), input.times)
        obs = input.obs
        self.assertEqual(2, obs.shape[0])  # Time
        self.assertEqual(3, obs.shape[1])  # Leadtime
        self.assertEqual(1, obs.shape[2])  # Location
        self.assertAlmostEqual(1, obs[0, 0, 0])
        self.assertAlmostEqual(2, obs[0, 1, 0])
        self.assertAlmostEqual(3, obs[0, 2, 0])
        self.assertAlmostEqual(4, obs[1, 0, 0])
        self.assertAlmostEqual(5, obs[1, 1, 0])
        self.assertAlmostEqual(6, obs[1, 2, 0])
        self.assertEqual(0, len(input.thresholds))
        self.assertEqual(0, len(input.quantiles))
        self.assertTrue(input.ensemble is not None)
        self.assertTrue(input.ensemble.shape[-1] == 0)
        self.assertTrue(verif.field.Obs() in input.get_fields())
        self.assertFalse(verif.field.Fcst() in input.get_fields())

    def test_is_valid(self):
        self.assertTrue(
            verif.input.Netcdf.is_valid("verif/tests/files/netcdf_valid1.nc")
        )
        self.assertTrue(
            verif.input.Netcdf.is_valid("verif/tests/files/netcdf_valid2.nc")
        )
        # Test that if altitude is missing, the file is still valid
        self.assertTrue(
            verif.input.Netcdf.is_valid("verif/tests/files/netcdf_valid3.nc")
        )
        verif.input.Netcdf("verif/tests/files/netcdf_valid1.nc")
        verif.input.Netcdf("verif/tests/files/netcdf_valid2.nc")
        verif.input.Netcdf("verif/tests/files/netcdf_valid3.nc")

    def test_invalid(self):
        self.assertFalse(
            verif.input.Comps.is_valid("verif/tests/files/netcdf_invalid1.nc")
        )
        self.assertFalse(
            verif.input.Comps.is_valid("verif/tests/files/netcdf_invalid2.nc")
        )

    def test_has(self):
        input = verif.input.Netcdf("verif/tests/files/netcdf_valid1.nc")
        self.assertTrue(input.has_field(verif.field.Obs()))
        self.assertTrue(input.has_field(verif.field.Fcst()))
        self.assertFalse(input.has_field(verif.field.Ensemble()))
        self.assertFalse(input.has_field(verif.field.EnsembleMean()))
        self.assertFalse(input.has_field(verif.field.EnsembleVariance()))
        self.assertFalse(input.has_field(verif.field.Threshold(1)))
        self.assertFalse(input.has_field(verif.field.Quantile(1)))

    def test_has2(self):
        input = verif.input.Netcdf("verif/tests/files/netcdf_valid2.nc")
        self.assertTrue(input.has_field(verif.field.Obs()))
        self.assertFalse(input.has_field(verif.field.Fcst()))
        self.assertFalse(input.has_field(verif.field.Ensemble()))
        self.assertFalse(input.has_field(verif.field.EnsembleMean()))
        self.assertFalse(input.has_field(verif.field.EnsembleVariance()))
        self.assertFalse(input.has_field(verif.field.Threshold(1)))
        self.assertFalse(input.has_field(verif.field.Quantile(1)))

    def test_has_ens(self):
        input = verif.input.Netcdf("verif/tests/files/netcdf_ens.nc")
        self.assertTrue(input.has_field(verif.field.Obs()))
        self.assertTrue(input.has_field(verif.field.Fcst()))
        self.assertTrue(input.has_field(verif.field.Ensemble()))
        self.assertTrue(input.has_field(verif.field.EnsembleMean()))
        self.assertFalse(input.has_field(verif.field.EnsembleVariance()))
        self.assertFalse(input.has_field(verif.field.Threshold(1)))
        self.assertFalse(input.has_field(verif.field.Quantile(1)))


if __name__ == "__main__":
    unittest.main()



import unittest
import verif.input
import verif.location
import verif.data


class MyTest(unittest.TestCase):
    def test_comps_to_verif_threshold(self):
        self.assertEqual(0, verif.input.Comps._comps_to_verif_threshold("p0"))
        self.assertEqual(0.1, verif.input.Comps._comps_to_verif_threshold("p01"))
        self.assertEqual(0.01, verif.input.Comps._comps_to_verif_threshold("p001"))
        self.assertEqual(2, verif.input.Comps._comps_to_verif_threshold("p2"))
        self.assertEqual(1.1, verif.input.Comps._comps_to_verif_threshold("p1.1"))
        self.assertEqual(25.1, verif.input.Comps._comps_to_verif_threshold("p25.1"))
        self.assertEqual(-2, verif.input.Comps._comps_to_verif_threshold("pm2"))
        self.assertEqual(-1.2, verif.input.Comps._comps_to_verif_threshold("pm1.2"))

        self.assertEqual(None, verif.input.Comps._comps_to_verif_threshold("qwoei"))
        self.assertEqual(None, verif.input.Comps._comps_to_verif_threshold("q0.r"))

    def test_comps_to_verif_quantile(self):
        self.assertEqual(0, verif.input.Comps._comps_to_verif_quantile("q0"))
        self.assertEqual(0.1, verif.input.Comps._comps_to_verif_quantile("q10"))
        self.assertEqual(0.01, verif.input.Comps._comps_to_verif_quantile("q1"))
        self.assertEqual(0.001, verif.input.Comps._comps_to_verif_quantile("q01"))
        self.assertEqual(1, verif.input.Comps._comps_to_verif_quantile("q100"))

        self.assertEqual(None, verif.input.Comps._comps_to_verif_quantile("q101"))
        self.assertEqual(None, verif.input.Comps._comps_to_verif_quantile("qm101"))

    def test_verif_to_comps_threshold(self):
        self.assertEqual("p0", verif.input.Comps._verif_to_comps_threshold(0))
        self.assertEqual("p01", verif.input.Comps._verif_to_comps_threshold(0.1))
        self.assertEqual("p1", verif.input.Comps._verif_to_comps_threshold(1))
        self.assertEqual("p10", verif.input.Comps._verif_to_comps_threshold(10))
        self.assertEqual("pm01", verif.input.Comps._verif_to_comps_threshold(-0.1))
        self.assertEqual("pm1", verif.input.Comps._verif_to_comps_threshold(-1))
        self.assertEqual("pm10", verif.input.Comps._verif_to_comps_threshold(-10))

    def test_verif_to_comps_quantile(self):
        self.assertEqual("q0", verif.input.Comps._verif_to_comps_quantile(0))
        self.assertEqual("q10", verif.input.Comps._verif_to_comps_quantile(0.1))
        self.assertEqual("q1", verif.input.Comps._verif_to_comps_quantile(0.01))
        self.assertEqual("q01", verif.input.Comps._verif_to_comps_quantile(0.001))
        self.assertEqual("q100", verif.input.Comps._verif_to_comps_quantile(1))

        self.assertEqual(None, verif.input.Comps._verif_to_comps_quantile(-1))
        self.assertEqual(None, verif.input.Comps._verif_to_comps_quantile(2))

    def test_valid(self):
        self.assertTrue(verif.input.Comps.is_valid("verif/tests/files/comps_valid1.nc"))
        self.assertTrue(verif.input.Comps.is_valid("verif/tests/files/comps_valid2.nc"))

    def test_invalid(self):
        self.assertFalse(
            verif.input.Comps.is_valid("verif/tests/files/comps_invalid1.nc")
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import verif.data
import numpy as np


def get_data_from_text(filename):
    data = verif.data.Data(verif.input.Text(filename))
    return data


def assert_set_equal(array1, array2):
    np.testing.assert_array_equal(np.sort(array1), np.sort(array2))


class TestData(unittest.TestCase):
    def test_doesnotexist(self):
        with self.assertRaises(SystemExit):
            inputs = [
                verif.input.get_input(filename)
                for filename in [
                    "verif/tests/files/file1.txt",
                    "verif/tests/files/doesnotexist.txt",
                ]
            ]
            data = verif.data.Data(inputs=inputs)

    def test_simple(self):
        inputs = [
            verif.input.get_input(filename)
            for filename in [
                "verif/tests/files/file1.txt",
                "verif/tests/files/file2.txt",
            ]
        ]
        data = verif.data.Data(inputs=inputs)
        lats = [loc.lat for loc in data.locations]
        self.assertTrue(len(lats) == 1)
        self.assertTrue(lats[0] == 42)
        lons = [loc.lon for loc in data.locations]
        self.assertTrue(len(lons) == 1)
        self.assertTrue(lons[0] == 23)

        leadtimes = data.leadtimes
        self.assertEqual(2, leadtimes.shape[0])
        self.assertEqual(0, leadtimes[0])
        self.assertEqual(12, leadtimes[1])

        times = data.times
        self.assertEqual(3, times.shape[0])
        # 20120101: Common leadtimes: 0 and 12, locations 41
        fields = [verif.field.Obs(), verif.field.Fcst()]
        axis = verif.axis.Time()

        [obs, fcst] = data.get_scores(fields, 0, axis, 0)
        self.assertEqual(5, fcst[0])  # Offset 0
        self.assertEqual(7, fcst[1])  # Offset 12
        self.assertEqual(-1, obs[0])
        self.assertEqual(2, obs[1])
        # 20120102: (missing obs at leadtime 12)
        [obs, fcst] = data.get_scores(fields, 0, axis, 1)
        self.assertEqual(1, fcst.shape[0])
        self.assertEqual(6, fcst[0])
        self.assertEqual(1, obs[0])

        # 20120101
        [obs, fcst] = data.get_scores(fields, 1, axis, 0)
        self.assertEqual(0, fcst[0])  # Offset 0
        self.assertEqual(1, fcst[1])  # Offset 12
        self.assertEqual(-1, obs[0])
        self.assertEqual(2, obs[1])
        # 20120102
        [obs, fcst] = data.get_scores(fields, 1, axis, 1)
        self.assertEqual(1, fcst.shape[0])
        self.assertEqual(5, fcst[0])
        self.assertEqual(1, obs[0])

    def test_latrange(self):
        inputs = [
            verif.input.Text("verif/tests/files/file1.txt"),
            verif.input.Text("verif/tests/files/file3.txt"),
        ]
        data = verif.data.Data(
            inputs, lat_range=[44, 60]
        )  # Only 1 common station within the range
        self.assertEqual(1, len(data.locations))
        self.assertEqual(50, data.locations[0].lat)
        self.assertEqual(10, data.locations[0].lon)
        data = verif.data.Data(inputs, lat_range=[40, 60])
        self.assertEqual(2, len(data.locations))
        with self.assertRaises(SystemExit):
            data = verif.data.Data(inputs, lat_range=[55, 60])

    def test_latrange_locations(self):
        # Check that the intersection of lat_range and locations is correct
        inputs = [verif.input.Text("verif/tests/files/file_many_locations.txt")]
        data = verif.data.Data(inputs, lat_range=[25, 60], locations=[3, 2, 21])
        self.assertEqual(2, len(data.locations))
        self.assertTrue(3 in [loc.id for loc in data.locations])
        self.assertTrue(21 in [loc.id for loc in data.locations])

    def test_lonrange_locations(self):
        # Check that the intersection of lon_range and locations is correct
        inputs = [verif.input.Text("verif/tests/files/file_many_locations.txt")]
        data = verif.data.Data(inputs, lon_range=[6, 16], locations=[1, 2, 911])
        self.assertEqual(2, len(data.locations))
        self.assertTrue(1 in [loc.id for loc in data.locations])
        self.assertTrue(2 in [loc.id for loc in data.locations])

    def test_obsrange(self):
        inputs = [
            verif.input.Text("verif/tests/files/file1.txt"),
            verif.input.Text("verif/tests/files/file3.txt"),
        ]
        data = verif.data.Data(inputs=inputs, obs_range=[0, 3.5])

        fields = [verif.field.Obs(), verif.field.Fcst()]
        axis = verif.axis.Time()

        [obs, fcst] = data.get_scores(fields, 0, axis, 0)
        self.assertEqual(6, fcst[0])  # Offset 0
        self.assertEqual(7, fcst[1])  # Offset 12
        self.assertEqual(3, obs[0])
        self.assertEqual(2, obs[1])
        # 20120102: (missing obs at leadtime 12)
        [obs, fcst] = data.get_scores(fields, 0, axis, 1)
        self.assertEqual(1, fcst.shape[0])
        self.assertEqual(6, fcst[0])
        self.assertEqual(1, obs[0])

    def test_dayofyear(self):
        inputs = [verif.input.get_input("verif/tests/files/file1.txt")]
        data = verif.data.Data(inputs=inputs)
        axis = verif.axis.Dayofyear()
        axis_values = data.get_axis_values(axis)
        self.assertEqual(3, len(axis_values))
        self.assertEqual(1, axis_values[0])
        self.assertEqual(2, axis_values[1])
        self.assertEqual(3, axis_values[2])

        # All the values for day 1
        values = data.get_scores([verif.field.Fcst()], 0, axis, 0)[0]
        self.assertEqual(6, len(values))
        assert_set_equal(np.array([6, 7, 7, 5, 4, 7]), values)

        values = data.get_scores([verif.field.Fcst()], 0, axis, 2)[0]
        self.assertEqual(5, len(values))
        assert_set_equal(np.array([-4, 3, 9, 12, 16]), values)


class TestRepeated(unittest.TestCase):
    """
    Test that having repeated times, leadtimes, or locations in a file doesn't cause runtime errors
    """

    def test_repeated_times(self):
        inputs = [verif.input.get_input("verif/tests/files/netcdf_repeated_times.nc")]
        data = verif.data.Data(inputs=inputs)
        self.assertEqual(2, len(data.times))
        self.assertEqual(3, len(data.leadtimes))
        self.assertEqual(1, len(data.locations))

        # Check that we only have two times
        obs = data.get_scores([verif.field.Obs()], 0, verif.axis.All())[0]
        np.testing.assert_array_equal(np.array([[1, 2, 3], [4, 5, 6]]), obs[:, :, 0])

        # Check that subset of times works
        data = verif.data.Data(inputs=inputs, times=inputs[0].times[[0, 2]])
        obs = data.get_scores([verif.field.Obs()], 0, verif.axis.All(), 0)[0]
        np.testing.assert_array_equal(np.array([[1, 2, 3]]), obs[:, :, 0])

    def test_repeated_leadtimes(self):
        inputs = [
            verif.input.get_input("verif/tests/files/netcdf_repeated_leadtimes.nc")
        ]
        data = verif.data.Data(inputs=inputs)
        self.assertEqual(2, len(data.times))
        self.assertEqual(2, len(data.leadtimes))
        self.assertEqual(1, len(data.locations))

        obs = data.get_scores([verif.field.Obs()], 0, verif.axis.All())[0]
        np.testing.assert_array_equal(np.array([[1, 2], [4, 5]]), obs[:, :, 0])

        # Check that subset of locations works
        data = verif.data.Data(inputs=inputs, leadtimes=inputs[0].leadtimes[[0, 2]])
        obs = data.get_scores([verif.field.Obs()], 0, verif.axis.All())[0]
        ar = np.zeros([2, 1])
        ar[0] = 1
        ar[1] = 4
        np.testing.assert_array_equal(ar, obs[:, :, 0])

    def test_repeated_locations(self):
        inputs = [
            verif.input.get_input("verif/tests/files/netcdf_repeated_locations.nc")
        ]
        data = verif.data.Data(inputs=inputs)
        self.assertEqual(2, len(data.times))
        self.assertEqual(2, len(data.leadtimes))
        self.assertEqual(2, len(data.locations))

        obs = data.get_scores([verif.field.Obs()], 0, verif.axis.All())[0]
        self.assertEqual(2, obs.shape[2])
        np.testing.assert_array_equal(np.array([[1, 4], [7, 10]]), obs[:, :, 0])
        np.testing.assert_array_equal(np.array([[2, 5], [8, 11]]), obs[:, :, 1])

        # Check that intersections with locations work
        data = verif.data.Data(inputs=inputs, locations=[18700, 18700])
        obs = data.get_scores([verif.field.Obs()], 0, verif.axis.All())[0]
        self.assertEqual(1, obs.shape[2])
        np.testing.assert_array_equal(np.array([[1, 4], [7, 10]]), obs[:, :, 0])

    def test_repeated_times_two_files(self):
        inputs = [
            verif.input.get_input("verif/tests/files/netcdf_repeated_%s.nc" % name)
            for name in ["times", "leadtimes"]
        ]
        data = verif.data.Data(inputs=inputs)
        self.assertEqual(2, len(data.times))
        self.assertEqual(2, len(data.leadtimes))
        self.assertEqual(1, len(data.locations))

        fcst = data.get_scores([verif.field.Fcst()], 0, verif.axis.All())[0]
        np.testing.assert_array_equal(np.array([[11, 12], [14, 15]]), fcst[:, :, 0])

        fcst = data.get_scores([verif.field.Fcst()], 1, verif.axis.All())[0]
        np.testing.assert_array_equal(np.array([[11, 12], [14, 15]]), fcst[:, :, 0])


class TestDataRemovingLocations(unittest.TestCase):
    def test_inside_lat_range(self):
        inputs = [verif.input.Text("verif/tests/files/file3locations.txt")]
        data = verif.data.Data(inputs, lat_range=[39, 51], locations_x=[2])
        self.assertEqual(1, len(data.locations))
        self.assertEqual(1, data.locations[0].id)
        self.assertEqual(40, data.locations[0].lat)

    def test_outside_lat_range(self):
        inputs = [verif.input.Text("verif/tests/files/file3locations.txt")]
        data = verif.data.Data(inputs, lat_range=[39, 51], locations_x=[3])
        self.assertEqual(2, len(data.locations))
        self.assertTrue(1 in [loc.id for loc in data.locations])
        self.assertTrue(2 in [loc.id for loc in data.locations])

    def test_with_l1(self):
        inputs = [verif.input.Text("verif/tests/files/file3locations.txt")]
        data = verif.data.Data(inputs, locations=[1, 2], locations_x=[2])
        self.assertEqual(1, len(data.locations))
        self.assertTrue(1 in [loc.id for loc in data.locations])

    def test_with_l2(self):
        inputs = [verif.input.Text("verif/tests/files/file3locations.txt")]
        data = verif.data.Data(inputs, locations=[1, 2], locations_x=[3])
        self.assertEqual(2, len(data.locations))
        self.assertTrue(1 in [loc.id for loc in data.locations])
        self.assertTrue(2 in [loc.id for loc in data.locations])


class TestDataRemovingTimes1(unittest.TestCase):
    def test_times(self):
        inputs = [verif.input.Text("verif/tests/files/file1.txt")]
        data = verif.data.Data(inputs, times=[1325376000, 1325548800])
        self.assertEqual(2, len(data.times))
        self.assertEqual(1325376000, data.times[0])
        self.assertEqual(1325548800, data.times[1])

    def test_dates(self):
        inputs = [verif.input.Text("verif/tests/files/file1.txt")]
        data = verif.data.Data(inputs, dates=[20120101])
        self.assertEqual(1, len(data.times))
        self.assertEqual(1325376000, data.times[0])


class TestDataRemovingTimes4(unittest.TestCase):
    """
    Check that various combinations of -t -d and -tod remove the correct times
    """

    def test_times(self):
        inputs = [verif.input.Text("verif/tests/files/file4.txt")]
        data = verif.data.Data(inputs, times=[1325376000, 1325548800])
        self.assertEqual(2, len(data.times))
        self.assertEqual(1325376000, data.times[0])
        self.assertEqual(1325548800, data.times[1])

    def test_dates(self):
        inputs = [verif.input.Text("verif/tests/files/file4.txt")]
        data = verif.data.Data(inputs, dates=[20120101, 20120102])
        self.assertEqual(3, len(data.times))
        self.assertEqual(1325376000, data.times[0])
        self.assertEqual(1325397600, data.times[1])
        self.assertEqual(1325462400, data.times[2])

    def test_times_dates(self):
        inputs = [verif.input.Text("verif/tests/files/file4.txt")]
        data = verif.data.Data(
            inputs, times=[1325376000, 1325548800], dates=[20120101, 20120102]
        )
        self.assertEqual(1, len(data.times))
        self.assertEqual(1325376000, data.times[0])

    def test_timeofday(self):
        inputs = [verif.input.Text("verif/tests/files/file4.txt")]
        data = verif.data.Data(inputs, tods=[6])
        self.assertEqual(2, len(data.times))
        self.assertEqual(1325397600, data.times[0])
        self.assertEqual(1325570400, data.times[1])

        # Check that data retrieved is of the correct size
        obs, fcst = data.get_scores([verif.field.Obs(), verif.field.Fcst()], 0)
        self.assertEqual(2, obs.shape[0])
        self.assertEqual(3, obs.shape[1])
        self.assertEqual(1, obs.shape[2])
        np.testing.assert_array_equal(np.array([[11, 4, 9], [3, 5, 6]]), obs[:, :, 0])
        np.testing.assert_array_equal(np.array([[6, 7, 7], [3, 1, 2]]), fcst[:, :, 0])

        data = verif.data.Data(inputs, tods=[0])
        self.assertEqual(3, len(data.times))
        self.assertEqual(1325376000, data.times[0])
        self.assertEqual(1325462400, data.times[1])
        self.assertEqual(1325548800, data.times[2])

    def test_times_dates_timeofday(self):
        inputs = [verif.input.Text("verif/tests/files/file4.txt")]
        data = verif.data.Data(
            inputs,
            times=[1325462400, 1325548800, 1325570400],
            dates=[20120101, 20120103],
            tods=[0],
        )
        self.assertEqual(1, len(data.times))
        self.assertEqual(1325548800, data.times[0])


class TestShareObs(unittest.TestCase):
    def test_reused(self):
        # Observations from file1 should be reused in file1_no_obs
        inputs = [
            verif.input.Text("verif/tests/files/file1.txt"),
            verif.input.Text("verif/tests/files/file1_no_obs.txt"),
        ]
        data = verif.data.Data(inputs)
        obs0 = data.get_scores(verif.field.Obs(), 0)
        obs1 = data.get_scores(verif.field.Obs(), 1)
        np.testing.assert_array_equal(obs0, obs1)

    def test_different_missing(self):
        # The two files are identical, other than some obs and fcst are missing for different times
        # The obs and fcst arrays after retriving through data should be the same
        inputs = [
            verif.input.Text("verif/tests/files/file1.txt"),
            verif.input.Text("verif/tests/files/file1_different_missing.txt"),
        ]
        data = verif.data.Data(inputs)
        fields = [verif.field.Obs(), verif.field.Fcst()]
        obs0, fcst0 = data.get_scores(fields, 0)
        obs1, fcst1 = data.get_scores(fields, 1)
        np.testing.assert_array_equal(obs0, obs1)
        np.testing.assert_array_equal(fcst0, fcst1)


class TestDataClim(unittest.TestCase):
    def test_names(self):
        """
        Checks that climatology files do not end up in get_names()
        and related functions
        """
        inputs = [verif.input.Text("verif/tests/files/file1.txt")]
        clim = verif.input.get_input("verif/tests/files/file2.txt")
        data = verif.data.Data(inputs, clim=clim)
        self.assertEqual(1, len(data.get_names()))
        self.assertEqual(1, len(data.get_short_names()))
        self.assertEqual(1, len(data.get_full_names()))


class TestDataFields(unittest.TestCase):
    def test_get_fields1(self):
        inputs = [verif.input.Text("verif/tests/files/file1.txt")]
        data = verif.data.Data(inputs)
        self.assertTrue(verif.field.Fcst() in data.get_fields())
        self.assertTrue(verif.field.Obs() in data.get_fields())

    def test_get_fields2(self):
        inputs = [verif.input.Text("verif/tests/files/file1_no_fcst.txt")]
        data = verif.data.Data(inputs)
        self.assertFalse(verif.field.Fcst() in data.get_fields())
        self.assertTrue(verif.field.Obs() in data.get_fields())

    def test_get_fields3(self):
        # Only Obs is common between the two files
        inputs = [
            verif.input.Text("verif/tests/files/%s" % file)
            for file in ["file1.txt", "file1_no_fcst.txt"]
        ]
        data = verif.data.Data(inputs)
        self.assertFalse(verif.field.Fcst() in data.get_fields())
        self.assertTrue(verif.field.Obs() in data.get_fields())


class TestDataNanTime(unittest.TestCase):
    def test(self):
        data = verif.data.Data(
            [verif.input.Text("verif/tests/files/file_nan_time.txt")]
        )
        np.testing.assert_array_equal(
            np.array([1325376000, 1325462400, 1325548800]), data.times
        )
        obs = data.get_scores(verif.field.Obs(), 0)


if __name__ == "__main__":
    unittest.main()



import unittest
import verif.interval
import numpy as np


class MyTest(unittest.TestCase):
    _tests = [-0.1, 0, 0.1, 0.9, 1, 1.1]

    def compare(self, interval, expected):
        """Run interval.within on all elements in _tests and compare the result with expected"""
        for i in range(0, len(self._tests)):
            self.assertEqual(expected[i], interval.within(self._tests[i]))

    def test_1(self):
        interval = verif.interval.Interval(0, 1, False, False)  # 0 < x < 1
        self.compare(interval, [False, False, True, True, False, False])

    def test_2(self):
        interval = verif.interval.Interval(0, 1, True, True)  # 0 <= x <= 1
        self.compare(interval, [False, True, True, True, True, False])

    def test_3(self):
        interval = verif.interval.Interval(0, 1, False, True)  # 0 < x <= 1
        self.compare(interval, [False, False, True, True, True, False])

    def test_4(self):
        interval = verif.interval.Interval(0, 1, True, False)  # 0 <= x < 1
        self.compare(interval, [False, True, True, True, False, False])

    def test_zero_interval_1(self):
        interval = verif.interval.Interval(0, 0, False, False)  # 0 < x < 0
        self.compare(interval, [False, False, False, False, False, False])

    def test_zero_interval_2(self):
        interval = verif.interval.Interval(0, 0, True, True)  # 0 <= x <= 0
        self.compare(interval, [False, True, False, False, False, False])

    def test_zero_interval_3(self):
        interval = verif.interval.Interval(0, 0, False, True)  # 0 < x <= 0
        self.compare(interval, [False, False, False, False, False, False])

    def test_zero_interval_4(self):
        interval = verif.interval.Interval(0, 0, True, False)  # 0 <= x < 0
        self.compare(interval, [False, False, False, False, False, False])

    def test_equal(self):
        from verif.interval import Interval

        self.assertEqual(Interval(0, 0, True, False), Interval(0, 0, True, False))
        self.assertTrue(Interval(0, 0, True, False) == Interval(0, 0, True, False))
        self.assertFalse(Interval(0, 0, True, False) != Interval(0, 0, True, False))

    def test_unequal(self):
        from verif.interval import Interval

        for interval in [
            Interval(0, 0, True, True),
            Interval(0, 0, False, False),
            Interval(0, 1, True, False),
            Interval(1, 0, True, False),
        ]:
            self.assertFalse(Interval(0, 0, True, False) == interval)
            self.assertFalse(Interval(0, 0, True, False) == interval)
            self.assertTrue(Interval(0, 0, True, False) != interval)
            self.assertTrue(Interval(0, 0, True, False) != interval)

    def test_array(self):
        ar = np.array([1, 3, 2, 0, 15])
        interval = verif.interval.Interval(2, 5, True, True)
        np.testing.assert_array_equal(
            np.array([False, True, True, False, False]), interval.within(ar)
        )
        np.testing.assert_array_equal(np.array(True), interval.within(3))

    def test_nan(self):
        """
        Test within returns np.nan when the input is np.nan. Interval.within
        returns a masked array, which is converted to a regular aray so that it
        can be tested here.
        """
        ar = np.array([1, 3, 2, np.nan, 15])
        interval = verif.interval.Interval(2, 5, True, True)
        values = interval.within(ar)
        values = np.ma.filled(values, fill_value=np.nan)
        answer = np.ma.masked_array(
            [False, True, True, np.nan, False], mask=[0, 0, 0, 1, 0]
        )
        np.testing.assert_array_equal(answer, values)

        # Test scalar
        self.assertTrue(np.isnan(interval.within(np.nan)))


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import verif
import verif.input
import verif.data
import verif.location


class InputTextTest(unittest.TestCase):
    def test_ensemble_mean(self):
        input = verif.input.Text("verif/tests/files/file1_ens.txt")
        print(input.ensemble[0, 0, 0, :])
        mean = input.ensemble_mean
        self.assertEqual((3, 3, 2, 3), input.ensemble.shape)
        self.assertEqual((3, 3, 2), mean.shape)

        np.testing.assert_array_equal((4, 4, 5), input.ensemble[0, 0, 0, :])
        self.assertAlmostEqual(4.333333333, mean[0, 0, 0])


if __name__ == "__main__":
    unittest.main()



import unittest
import verif.data
import numpy as np


def get_data_from_text(filename):
    data = verif.data.Data(verif.input.Text(filename))
    return data


def assert_set_equal(array1, array2):
    np.testing.assert_array_equal(np.sort(array1), np.sort(array2))


class TestData(unittest.TestCase):
    def test_simple(self):
        inputs = [
            verif.input.get_input(filename)
            for filename in [
                "verif/tests/files/file2_ens.txt",
                "verif/tests/files/file3_ens.txt",
            ]
        ]
        data = verif.data.Data(inputs=inputs)

        fields = [verif.field.Ensemble()]
        axis = verif.axis.Leadtime()

        # 3 members
        (ens,) = data.get_scores(fields, 0, axis, 0)
        self.assertEqual((3, 3), ens.shape)
        np.testing.assert_array_almost_equal([3, 4, 5], ens[0, :])

        # 2 members
        (ens,) = data.get_scores(fields, 1, axis, 0)
        self.assertEqual((3, 2), ens.shape)
        np.testing.assert_array_almost_equal([1, 2], ens[0, :])

    def test_preaggregated(self):
        inputs = [
            verif.input.get_input(filename)
            for filename in [
                "verif/tests/files/file2_ens.txt",
                "verif/tests/files/file3_ens.txt",
            ]
        ]
        data = verif.data.Data(
            inputs=inputs, dim_agg_length=12, dim_agg_method=verif.aggregator.Sum()
        )

        fields = [verif.field.Ensemble()]
        axis = verif.axis.Leadtime()

        # Look for the sum of leadtime 0 and 1
        leadtime = 1

        # 3 members
        (ens,) = data.get_scores(fields, 0, axis, leadtime)
        self.assertEqual((3, 3), ens.shape)
        np.testing.assert_array_almost_equal([7, 8, 9], ens[0, :])

        # 2 members
        (ens,) = data.get_scores(fields, 1, axis, leadtime)
        self.assertEqual((3, 2), ens.shape)
        np.testing.assert_array_almost_equal([4, 6], ens[0, :])


if __name__ == "__main__":
    unittest.main()



import unittest
from verif.location import Location


class TestLocation(unittest.TestCase):
    def test_equal(self):
        self.assertEqual(Location(0, 1, 2, 3), Location(0, 1, 2, 3))
        self.assertTrue(Location(0, 1, 2, 3) == Location(0, 1, 2, 3))
        self.assertFalse(Location(0, 1, 2, 3) != Location(0, 1, 2, 3))

    def test_unequal(self):
        for location in [
            Location(0, 1, 2, 2),
            Location(0, 1, 3, 3),
            Location(0, 2, 2, 3),
            Location(1, 1, 2, 3),
        ]:
            self.assertFalse(Location(0, 1, 2, 3) == location)
            self.assertFalse(Location(0, 1, 2, 3) == location)
            self.assertTrue(Location(0, 1, 2, 3) != location)
            self.assertTrue(Location(0, 1, 2, 3) != location)

    def test_distance(self):
        loc1 = Location(0, lat=60, lon=10, elev=0)
        loc2 = Location(0, lat=-30, lon=-40, elev=100)
        self.assertEqual(0, loc1.get_distance(loc1))
        # Computed using: http://www.onlineconversion.com/map_greatcircle_distance.htm
        self.assertAlmostEqual(10996966.188406856, loc1.get_distance(loc2))
        self.assertAlmostEqual(loc2.get_distance(loc1), loc1.get_distance(loc2))


if __name__ == "__main__":
    unittest.main()



import argparse
import copy
import netCDF4
import numpy as np
import sys
import time as timing


import verif.util


removeMissing = True
nc_missing = netCDF4.default_fillvals["f4"]


def progress_bar(fraction, width, text=""):
    num_x = int(fraction * (width-len(text) - 2))
    num_space = width - num_x - len(text) - 2
    sys.stdout.write("\r" + text + "[" + "X" * num_x + " " * num_space + "]")
    sys.stdout.flush()


def main():
    parser = argparse.ArgumentParser(description='Expands the observations in a verif file to match desired leadtimes and initialization times ')
    parser.add_argument('file', help='Verif file')
    parser.add_argument('-o', metavar="FILE", help='Output verif file', dest="ofilename", required=True)
    parser.add_argument('-i', type=verif.util.parse_numbers, default=[0], help='Initialization times in UTC (hours)', dest="init_times")
    parser.add_argument('-lt', type=verif.util.parse_numbers, help='Lead times (hours)', dest="lead_times")
    parser.add_argument('-t', type=verif.util.parse_numbers, help='Thresholds', dest="thresholds")
    parser.add_argument('-q', type=verif.util.parse_numbers, help='Quantiles', dest="quantiles")
    parser.add_argument('--debug', help='Display debug information', action="store_true")

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()

    if args.debug:
        width = verif.util.get_screen_width()

    input = verif.input.get_input(args.file)
    itimes = input.times
    ileadtimes = input.leadtimes
    locations = input.locations
    variable = input.variable

    itimes_whole_days = np.unique(np.array([int(t / 86400) * 86400 for t in input.times]))
    otimes = np.zeros(len(itimes_whole_days) * len(args.init_times))
    for i in range(len(args.init_times)):
        I = range(i * len(itimes_whole_days), (i+1) * len(itimes_whole_days))
        otimes[I] = itimes_whole_days + args.init_times[i] * 3600
    oleadtimes = args.lead_times
    obs = nc_missing * np.ones([len(otimes), len(oleadtimes), len(locations)], float)
    shape = input.obs.shape
    allobs = np.reshape(input.obs, [shape[0]*shape[1], shape[2]])
    q, w = np.meshgrid(ileadtimes, itimes)
    alltimes = q * 3600 + w

    for t in range(len(otimes)):
        if args.debug and (len(otimes) < width or t % int(len(otimes)/width) == 0):
            progress_bar(t*1.0 / len(otimes), width)

        for lt in range(len(oleadtimes)):
            I = np.where(alltimes.flatten() == otimes[t] + oleadtimes[lt] * 3600)[0]
            if len(I) >= 1:
                obs[t, lt, :] = allobs[I[0], :]

    output = netCDF4.Dataset(args.ofilename, 'w')
    output.createDimension("time", None)
    output.createDimension("leadtime", len(oleadtimes))
    output.createDimension("location", len(locations))
    if args.thresholds is not None:
        output.createDimension("threshold", len(args.thresholds))
        var = output.createVariable("threshold", "f4", ("threshold"))
        var[:] = args.thresholds
        output.createVariable("cdf", "f4", ("time", "leadtime", "location", "threshold"))
    if args.quantiles is not None:
        output.createDimension("quantile", len(args.quantiles))
        var = output.createVariable("quantile", "f4", ("quantile"))
        var[:] = args.quantiles
        output.createVariable("x", "f4", ("time", "leadtime", "location", "quantile"))
    vTime = output.createVariable("time", "i4", ("time",))
    vOffset = output.createVariable("leadtime", "f4", ("leadtime",))
    vLocation = output.createVariable("location", "i4", ("location",))
    vLat = output.createVariable("lat", "f4", ("location",))
    vLon = output.createVariable("lon", "f4", ("location",))
    vElev = output.createVariable("altitude", "f4", ("location",))
    vfcst = output.createVariable("fcst", "f4", ("time", "leadtime", "location"))
    vobs = output.createVariable("obs", "f4", ("time", "leadtime", "location"))
    output.standard_name = variable.name
    output.units = unit = variable.units.replace("$", "")

    vobs[:] = obs
    vTime[:] = otimes
    vOffset[:] = oleadtimes
    vLocation[:] = [s.id for s in locations]
    vLat[:] = [s.lat for s in locations]
    vLon[:] = [s.lon for s in locations]
    vElev[:] = [s.elev for s in locations]
    output.Conventions = "verif_1.0.0"
    output.close()


if __name__ == '__main__':
    main()



import unittest
import os
import sys
import numpy as np

np.seterr("raise")


class ScriptTest(unittest.TestCase):
    """
    These tests run verif on the command-line, but do not test the validity of the
    output graphics, only that they do or do not create errors.
    """

    @staticmethod
    def run_python(command):
        """Runs a verif command line"""
        command = "%s %s" % (sys.executable, command)
        status = os.system(command)
        if status != 0:
            raise Exception("Command failed: %s" % command)
        return

    def test_valid(self):
        programs = ["text2nc", "accumulate", "ens2prob", "expandverif", "window"]
        for program in programs:
            self.run_python("scripts/%s.py" % program)

    def test_invalid(self):
        """
        with self.assertRaises(SystemExit):
            self.run_with_image("verif --list-thresholds")
        with self.assertRaises(SystemExit):
            self.run_with_image("verif --list-quantiles")
        with self.assertRaises(SystemExit):
            self.run_with_image("verif --list-times")
        """


if __name__ == "__main__":
    unittest.main()



import unittest
import verif.axis
import verif.data
import verif.metric
import verif.output
import numpy as np


class TestOutput(unittest.TestCase):
    def test_text(self):
        inputs = [verif.input.get_input("verif/tests/files/file_sigdig.txt")]
        data = verif.data.Data(inputs=inputs)
        metric = verif.metric.Obs()

        output = verif.output.Standard(metric)
        output.text(data)

        output.axis = verif.axis.Time()
        output.text(data)


class TestStandard(unittest.TestCase):
    def _get(self):
        inputs = [
            verif.input.get_input(file)
            for file in ["examples/raw.txt", "examples/kf.txt"]
        ]
        data = verif.data.Data(inputs=inputs)
        metric = verif.metric.Mae()
        output = verif.output.Standard(metric)
        return data, output

    def test_leadtime(self):
        data, output = self._get()
        x, y, _, _, _ = output._get_x_y(data, verif.axis.Leadtime())
        lt = data.leadtimes
        np.testing.assert_array_equal(x, lt)
        self.assertEqual(2, y.shape[1])
        self.assertEqual(len(lt), y.shape[0])

    def test_location(self):
        data, output = self._get()
        for ax in [
            verif.axis.Location(),
            verif.axis.Lat(),
            verif.axis.Lon(),
            verif.axis.Elev(),
        ]:
            x, y, _, _, _ = output._get_x_y(data, verif.axis.Location())
            loc = data.locations
            self.assertEqual(len(loc), len(x))
            self.assertEqual(2, y.shape[1])
            self.assertEqual(len(loc), y.shape[0])

    def test_year(self):
        data, output = self._get()
        x, y, _, _, _ = output._get_x_y(data, verif.axis.Year())
        # There should only be one year in this dataset
        self.assertEqual(1, len(x))
        self.assertEqual(1, y.shape[0])
        # The date represents the first day of the year:
        self.assertEqual(20120101, verif.util.datenum_to_date(x[0]))
        self.assertEqual(2, y.shape[1])

    def test_month(self):
        data, output = self._get()
        x, y, _, _, _ = output._get_x_y(data, verif.axis.Month())
        # There should be three months in this dataset
        self.assertEqual(3, len(x))
        self.assertEqual(3, y.shape[0])
        # The dates represents the first day of the months:
        self.assertEqual(20120101, verif.util.datenum_to_date(x[0]))
        self.assertEqual(20120201, verif.util.datenum_to_date(x[1]))
        self.assertEqual(20120301, verif.util.datenum_to_date(x[2]))
        self.assertEqual(2, y.shape[1])

    def test_time(self):
        data, output = self._get()
        x, y, _, _, _ = output._get_x_y(data, verif.axis.Time())
        # 20120101 to 20120301
        self.assertEqual(61, len(x))
        self.assertEqual(61, y.shape[0])
        self.assertEqual(20120101, verif.util.datenum_to_date(x[0]))
        self.assertEqual(20120102, verif.util.datenum_to_date(x[1]))
        self.assertEqual(20120301, verif.util.datenum_to_date(x[-1]))
        self.assertEqual(2, y.shape[1])

    def test_no(self):
        # Collapse all data into one number
        data, output = self._get()
        x, y, _, _, _ = output._get_x_y(data, verif.axis.No())
        self.assertEqual(1, len(x))
        self.assertEqual(1, y.shape[0])
        self.assertEqual(2, y.shape[1])


class NameTest(unittest.TestCase):
    def test_valid(self):
        outputs = [x[1] for x in verif.output.get_all() if x[1].is_valid]
        for output in outputs:
            self.assertTrue(output.name != "")
            self.assertTrue(output.description != "")


if __name__ == "__main__":
    unittest.main()



import inspect
from . import metric_type
import numpy as np
import sys
import scipy.stats
import verif.aggregator
import verif.axis
import verif.interval
import verif.util


def get_all():
    """
    Returns a dictionary of all metric classes where the key is the class
    name (string) and the value is the class object
    """
    temp = inspect.getmembers(sys.modules[__name__], inspect.isclass)
    return temp


def get_all_by_type(type):
    """
    Like get_all, except only return metrics that are of a cerrtain
    verif.metric_type
    """
    temp = [m for m in get_all() if m[1].type == type]
    return temp


def get_all_obs_fcst_based():
    """ Like get_all, except only return obs-fcst-based metric classes """
    metrics = [metric for metric in get_all() if issubclass(metric[1], verif.metric.ObsFcstBased)]
    return metrics


def get(name):
    """ Returns an instance of an object with the given class name """
    metrics = get_all()
    m = None
    for metric in metrics:
        if name == metric[0].lower() and metric[1].is_valid():
            m = metric[1]()
    return m


def get_p(data, input_index, axis, axis_index, interval):
    """
    Retrieves and computes forecast probability and verifying observation for
    being inside interval

    Returns:
       obs (np.array): True when observation is inside interval
       p (np.array): True when forecast is inside interval

    """
    p0 = 0
    p1 = 1
    if interval.lower != -np.inf and interval.upper != np.inf:
        var0 = verif.field.Threshold(interval.lower)
        var1 = verif.field.Threshold(interval.upper)
        [obs, p0, p1] = data.get_scores([verif.field.Obs(), var0, var1],
              input_index, axis, axis_index)
    elif interval.lower != -np.inf:
        var0 = verif.field.Threshold(interval.lower)
        [obs, p0] = data.get_scores([verif.field.Obs(), var0], input_index,
              axis, axis_index)
    elif interval.upper != np.inf:
        var1 = verif.field.Threshold(interval.upper)
        [obs, p1] = data.get_scores([verif.field.Obs(), var1], input_index,
              axis, axis_index)

    obsP = interval.within(obs)
    obsP = np.ma.filled(obsP, fill_value=np.nan)
    p = p1 - p0  # Prob of obs within range
    return [obsP, p]


def get_q(data, input_index, axis, axis_index, interval):
    """
    Retrieve forecast quantile and verifying observation

    Returns:
       obs (np.array): True when observation is inside interval
       p (np.array): True when forecast is inside interval

    """
    p0 = 0
    p1 = 1
    var = verif.field.Quantile(interval.lower)
    [obs, q] = data.get_scores([verif.field.Obs(), var], input_index, axis, axis_index)

    return [obs, q]


class Metric(object):
    """ Class to compute a score for a verification metric

    Scores are computed by retrieving information from a verif.data.Data object.
    As data is organized in multiple dimensions, scores are computed for a
    particular verif.axis.Axis. Also data objects have several input files, so
    scores are computed for a particular input.

    The ObsFcstBased class offers a simple way to design a metric that only
    uses observations and forecasts from data.

    Class attributes:
       description (str): A short one-liner describing the metric. This will show
          up in the main verif documentation.
       long (str): A longer description. This will show up in the
          documentation when a specific metric is chosen.
       min (float): Minimum possible value the metric can take on. None if no min.
       max (float): Maximum possible value the metric can take on. None if no max.
       require_threshold_type (str) : What type of thresholds does this metric
          require? One of 'None', 'deterministic', 'threshold', 'quantile'.
       supports_threshold (bool) : Does it make sense to use '-x threshold' with this metric?
       supports_field (bool) : Does it make sense to use '-x obs' or '-x fcst' with this metric?
       orientation (int): 1 for a positively oriented score (higher values are better),
          -1 for negative, and 0 for all others
       reference (str): A string with an academic reference
       supports_aggregator: Does this metric use self.aggregator?
       type (verif.metric_type.MetricType): What type of metric is this?

    To implement a new metric:
       Fill out cls.description and implement compute_core(). The other class
       attributes (see above) are optional.
    """
    # This must be overloaded
    description = None

    # Default values
    long = None
    reference = None
    orientation = 0
    min = None
    max = None
    default_axis = verif.axis.Leadtime()  # If no axis is specified, use this axis as default
    default_bin_type = None
    require_threshold_type = None
    supports_threshold = False
    supports_field = False
    perfect_score = None
    aggregator = verif.aggregator.Mean()
    supports_aggregator = False
    type = verif.metric_type.Deterministic()
    min_num_thresholds = None
    max_num_thresholds = None

    def compute(self, data, input_index, axis, interval):
        """ Compute the score along an axis

        Arguments:
           data (verif.data.Data): data object to get information from
           input_index (int): input index to compute the result for
           axis (verif.axis.Axis): Axis to compute score for for
           interval: Compute score for this interval (only applies to some metrics)

        Returns:
           np.array: A 1D numpy array of one score for each slice along axis
        """
        size = data.get_axis_size(axis)
        scores = np.zeros(size, 'float')

        # Loop through axis indices
        for axis_index in range(0, size):
            x = self.compute_single(data, input_index, axis, axis_index, interval)
            scores[axis_index] = x
        return scores

    def compute_single(self, data, input_index, axis, axis_index, interval):
        """ Computes the score for a given slice

        Arguments:
           data (verif.data.Data): data object to get information from
           input_index (int): input index to compute the result for
           axis (verif.axis.Axis): Axis to compute score for for
           axis_index (int): Slice along the axis
           interval: Compute score for this interval (only applies to some metrics)

        Returns:
           float: Value representing the score for the slice
        """
        raise NotImplementedError()

    def label(self, variable):
        """ What is an appropriate y-axis label for this metric? Override this if
        the metric does not have the same units as the forecast variable """
        return self.name + " (" + variable.units + ")"

    class ClassProperty(property):
        def __get__(self, cls, owner):
            return self.fget.__get__(None, owner)()

    @ClassProperty
    @classmethod
    def name(cls):
        """ Use the class name as default
        """
        return cls.get_class_name()

    @classmethod
    def is_valid(cls):
        """ Is this a valid metric that can be initialized? """
        return cls.description is not None

    @classmethod
    def help(cls):
        s = ""
        if cls.description is not None:
            s = cls.description
        if cls.orientation != 0:
            s = s + "\n" + verif.util.green("Orientation: ")
            if cls.orientation == 1:
                s = s + "Positive (higher values are better)"
            elif cls.orientation == -1:
                s = s + "Negative (lower values are better)"
            else:
                s = s + "None"
        if cls.perfect_score is not None:
            s = s + "\n" + verif.util.green("Perfect score: ") + str(cls.perfect_score)
        if cls.min is not None:
            s = s + "\n" + verif.util.green("Minimum value: ") + str(cls.min)
        if cls.max is not None:
            s = s + "\n" + verif.util.green("Maximum value: ") + str(cls.max)
        if cls.long is not None:
            s = s + "\n" + verif.util.green("Description: ") + cls.long
        if cls.reference is not None:
            s = s + "\n" + verif.util.green("Reference: ") + cls.reference
        return s

    @classmethod
    def get_class_name(cls):
        name = cls.__name__
        return name


class ObsFcstBased(Metric):
    """ Class for scores that are based on observations and deterministic forecasts only """
    type = verif.metric_type.Deterministic()
    supports_field = True

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obs, fcst] = data.get_scores([verif.field.Obs(), verif.field.Fcst()], input_index, axis, axis_index)
        assert(obs.shape[0] == fcst.shape[0])
        if axis == verif.axis.Obs():
            I = np.where(interval.within(obs))
            obs = obs[I]
            fcst = fcst[I]
        elif axis == verif.axis.Fcst():
            I = np.where(interval.within(fcst))
            obs = obs[I]
            fcst = fcst[I]

        return self.compute_from_obs_fcst(obs, fcst, interval)

    def compute_from_obs_fcst(self, obs, fcst, interval=None):
        """ Compute the score using only the observations and forecasts

        obs and fcst must have the same length, but may contain nan values

        Arguments:
           obs (np.array): 1D array of observations
           fcst (np.array): 1D array of forecasts

        Returns:
           float: Value of score
        """

        # Remove missing values
        I = np.where((np.isnan(obs) | np.isnan(fcst)) == 0)[0]
        obs = obs[I]
        fcst = fcst[I]

        if obs.shape[0] > 0:
            return self._compute_from_obs_fcst(obs, fcst)
        else:
            return np.nan

    def _compute_from_obs_fcst(self, obs, fcst):
        """ Compute the score

        Obs and fcst are guaranteed to:
            - have the same length
            - length >= 1
            - no missing values
         """
        raise NotImplementedError()


class FromField(Metric):
    supports_aggregator = True
    supports_field = True

    def __init__(self, field, aux=None):
        """ Compute scores from a field

        Arguments:
           field (verif.field.field): Retrive data from this field
           aux (verif.field.Field): When reading field, also pull values for
              this field to ensure only common data points are returned
        """
        self._field = field
        self._aux = aux

    def compute_single(self, data, input_index, axis, axis_index, interval):
        fields = [self._field]
        axis_pos = None
        if axis == verif.axis.Obs():
            if self._field != verif.field.Obs():
                fields += [verif.field.Obs()]
            axis_pos = len(fields) - 1
        elif axis == verif.axis.Fcst():
            if self._field != verif.field.Fcst():
                fields += [verif.field.Fcst()]
            axis_pos = len(fields) - 1
        if self._aux is not None:
            fields += [self._aux]
        values_array = data.get_scores(fields, input_index, axis, axis_index)
        values = values_array[0]

        # Subset if we have a subsetting axis
        if axis_pos is not None:
            I = np.where(interval.within(values_array[axis_pos]))[0]
            values = values[I]

        return self.aggregator(values)

    def label(self, variable):
        return self.aggregator.name().title() + " of " + self._field.name()


class Obs(FromField):
    """ Retrives the observation

    Note: This cannot be a subclass of ObsFcstBased, since we don't want
    to remove obs for which the forecasts are missing. Same for Fcst.
    """
    type = verif.metric_type.Deterministic()
    name = "Observation"
    description = "Observed value"
    supports_aggregator = True
    orientation = 0

    def __init__(self):
        super(Obs, self).__init__(verif.field.Obs())

    def label(self, variable):
        return self.aggregator.name().title() + " of observation (" + variable.units + ")"


class Fcst(FromField):
    type = verif.metric_type.Deterministic()
    name = "Forecast"
    description = "Forecasted value"
    supports_aggregator = True
    orientation = 0

    def __init__(self):
        super(Fcst, self).__init__(verif.field.Fcst())

    def label(self, variable):
        return self.aggregator.name().title() + " of forecast (" + variable.units + ")"


class Mae(ObsFcstBased):
    description = "Mean absolute error"
    min = 0
    perfect_score = 0
    supports_aggregator = True
    orientation = -1
    name = "Mean absolute error"

    def _compute_from_obs_fcst(self, obs, fcst):
        return self.aggregator(abs(obs - fcst))

    def label(self, variable):
        return "MAE (" + variable.units + ")"


class Bias(ObsFcstBased):
    name = "Bias"
    description = "Bias (forecast - observation)"
    perfect_score = 0
    supports_aggregator = True
    orientation = 0

    def _compute_from_obs_fcst(self, obs, fcst):
        return self.aggregator(fcst - obs)


class Diff(ObsFcstBased):
    name = "Diff"
    description = "Difference in aggregated statistics (agg(forecast) - agg(observation))"
    perfect_score = 0
    supports_aggregator = True
    orientation = 0

    def _compute_from_obs_fcst(self, obs, fcst):
        return self.aggregator(fcst) - self.aggregator(obs)


class Ratio(ObsFcstBased):
    name = "Ratio"
    description = "Ratio of aggregated statistics (agg(forecast) / agg(observation))"
    perfect_score = 1
    supports_aggregator = True
    orientation = 0

    def _compute_from_obs_fcst(self, obs, fcst):
        num = self.aggregator(fcst)
        denum = self.aggregator(obs)

        if denum == 0:
            return np.nan
        return num / denum

    def label(self, variable):
        return "Ratio"


class Ef(ObsFcstBased):
    name = "Exceedance fraction"
    description = "Exeedance fraction: fraction of times that forecasts > observations"
    min = 0
    max = 1
    perfect_score = 0.5
    orientation = 0

    def _compute_from_obs_fcst(self, obs, fcst):
        Nfcst = np.sum(obs < fcst)
        return Nfcst / 1.0 / len(fcst)

    def label(self, variable):
        return "Fraction fcst > obs"


class StdError(ObsFcstBased):
    name = "Standard error"
    description = "Standard error (i.e. RMSE if forecast had no bias)"
    min = 0
    perfect_score = 0
    orientation = -1

    def _compute_from_obs_fcst(self, obs, fcst):
        bias = np.mean(obs - fcst)
        return np.mean((obs - fcst - bias) ** 2) ** 0.5


class ObsStdDev(ObsFcstBased):
    name = "Observation Standard Deviation"
    description = "Standard deviation calculated for observations"
    min = 0
    perfect_score = 0
    orientation = -1

    def _compute_from_obs_fcst(self, obs, fcst):
        return np.std(obs)


class FcstStdDev(ObsFcstBased):
    name = "Forecast Standard Deviation"
    description = "Standard deviation calculated for forecasts"
    min = 0
    perfect_score = 0
    orientation = -1

    def _compute_from_obs_fcst(self, obs, fcst):
        return np.std(fcst)


class Rmse(ObsFcstBased):
    name = "Root mean squared error"
    description = "Root mean squared error"
    min = 0
    perfect_score = 0
    supports_aggregator = True
    orientation = -1

    def _compute_from_obs_fcst(self, obs, fcst):
        return self.aggregator((obs - fcst) ** 2) ** 0.5

    def label(self, variable):
        return "RMSE (" + variable.units + ")"


class Rmsf(ObsFcstBased):
    name = "Root mean squared factor"
    description = "Root mean squared factor"
    min = 0
    perfect_score = 1
    supports_aggregator = True
    orientation = 0

    def _compute_from_obs_fcst(self, obs, fcst):
        return np.exp(self.aggregator((np.log(fcst / obs)) ** 2) ** 0.5)

    def label(self, variable):
        return "RMSF (" + variable.units + ")"


class Cmae(ObsFcstBased):
    name = "Cube-root mean absolute cubic error"
    description = "Cube-root mean absolute cubic error"
    min = 0
    perfect_score = 0
    supports_aggregator = True
    orientation = -1

    def _compute_from_obs_fcst(self, obs, fcst):
        return (self.aggregator(abs(obs ** 3 - fcst ** 3))) ** (1.0 / 3)

    def label(self, variable):
        return "CMAE (" + variable.units + ")"


class Nsec(ObsFcstBased):
    name = "Nash-Sutcliffe efficiency coefficient"
    description = "Nash-Sutcliffe efficiency coefficient"
    max = 1
    perfect_score = 1
    orientation = 1

    def _compute_from_obs_fcst(self, obs, fcst):
        meanobs = np.mean(obs)
        num = np.sum((fcst - obs) ** 2)
        denom = np.sum((obs - meanobs) ** 2)
        if denom == 0:
            return np.nan
        else:
            return 1 - num / denom

    def label(self, variable):
        return "NSEC"


class Nnsec(ObsFcstBased):
    name = "Normalized Nash-Sutcliffe efficiency coefficient"
    description = "Normalized Nash-Sutcliffe efficiency coefficient"
    min = 0
    max = 1
    perfect_score = 1
    orientation = 1

    def _compute_from_obs_fcst(self, obs, fcst):
        meanobs = np.mean(obs)
        num = np.sum((fcst - obs) ** 2)
        denom = np.sum((obs - meanobs) ** 2)
        if denom == 0:
            return np.nan
        else:
            nsec = 1 - num / denom
            return 1 / (2 - nsec)

    def label(self, variable):
        return "NNSEC"


class Kge(ObsFcstBased):
    name = "Kling-Gupta efficiency"
    description = "Kling-Gupta efficiency"
    max = 1
    perfect_score = 1
    orientation = 1

    def _compute_from_obs_fcst(self, obs, fcst):
        # Based on https://agrimetsoft.com/calculators/Kling-Gupta%20efficiency
        meanobs = np.mean(obs)
        meanfcst = np.mean(fcst)
        stdobs = np.std(obs)
        stdfcst = np.std(fcst)
        if stdobs == 0 or stdfcst == 0:
            return np.nan

        corr = np.corrcoef(obs, fcst)[1, 0]
        if np.nan in [stdobs, stdfcst, corr]:
            return np.nan
        else:
            c = corr - 1
            m = meanfcst / meanobs - 1
            s = stdfcst / stdobs - 1
            return 1 - np.sqrt(c**2 + m**2 + s**2)

    def label(self, variable):
        return "KGE"


class Alphaindex(ObsFcstBased):
    name = "Alpha index"
    description = "Alpha index"
    perfect_score = 0
    orientation = -1
    max = 2
    min = 0

    def _compute_from_obs_fcst(self, obs, fcst):
        meanobs = np.mean(obs)
        meanfcst = np.mean(fcst)
        num = np.sum((fcst - obs - meanfcst + meanobs) ** 2)
        denom = np.sum((fcst - meanfcst) ** 2 + (obs - meanobs) ** 2)
        if denom == 0:
            return np.nan
        else:
            return 1 - num / denom

    def label(self, variable):
        return self.name


class Leps(ObsFcstBased):
    name = "Linear error in probability space"
    description = "Linear error in probability space"
    min = 0
    perfect_score = 0
    orientation = -1

    def _compute_from_obs_fcst(self, obs, fcst):
        N = len(obs)
        # Compute obs quantiles
        Iobs = np.array(np.argsort(obs), 'float')
        qobs = Iobs / N

        # Compute the quantiles that the forecasts are relative
        # to the observations
        qfcst = np.zeros(N, 'float')
        sortobs = np.sort(obs)
        for i in range(0, N):
            I = np.where(fcst[i] < sortobs)[0]
            if len(I > 0):
                qfcst[i] = float(I[0]) / N
            else:
                qfcst[i] = 1
        return np.mean(abs(qfcst - qobs))

    def label(self, variable):
        return "LEPS"


class Dmb(ObsFcstBased):
    name = "Degree of mass balance"
    description = "Degree of mass balance (obs/fcst)"
    perfect_score = 1
    orientation = 0

    def _compute_from_obs_fcst(self, obs, fcst):
        return np.mean(obs) / np.mean(fcst)

    def label(self, variable):
        return self.description


class Mbias(ObsFcstBased):
    name = "Multiplicative bias"
    description = "Multiplicative bias (fcst/obs)"
    perfect_score = 1
    orientation = 0

    def _compute_from_obs_fcst(self, obs, fcst):
        num = np.nanmean(fcst)
        denum = np.nanmean(obs)

        if denum == 0:
            return np.nan
        return num / denum

    def label(self, variable):
        return self.description


class Corr(ObsFcstBased):
    name = "Correlation"
    description = "Correlation between observations and forecasts"
    min = 0  # Technically -1, but values below 0 are not as interesting
    max = 1
    perfect_score = 1
    orientation = 1

    def _compute_from_obs_fcst(self, obs, fcst):
        if len(obs) <= 1:
            return np.nan
        if np.var(fcst) == 0:
            return np.nan
        return np.corrcoef(obs, fcst)[1, 0]

    def label(self, variable):
        return self.name


class RankCorr(ObsFcstBased):
    name = "Rank correlation"
    description = "Rank correlation between observations and forecasts"
    min = 0  # Technically -1, but values below 0 are not as interesting
    max = 1
    perfect_score = 1
    orientation = 1

    def _compute_from_obs_fcst(self, obs, fcst):
        if len(obs) <= 1:
            return np.nan
        return scipy.stats.spearmanr(obs, fcst)[0]

    def label(self, variable):
        return self.name


class KendallCorr(ObsFcstBased):
    name = "Kendall correlation"
    description = "Kendall correlation between observations and forecasts"
    min = 0  # Technically -1, but values below 0 are not as interesting
    max = 1
    perfect_score = 1
    orientation = 1

    def _compute_from_obs_fcst(self, obs, fcst):
        if len(obs) <= 1:
            return np.nan
        if np.var(fcst) == 0:
            return np.nan
        return scipy.stats.kendalltau(obs, fcst)[0]

    def label(self, variable):
        return self.name


class DError(ObsFcstBased):
    name = "Distribution Error"
    description = "Distribution error"
    min = 0
    perfect_score = 0
    supports_aggregator = False
    orientation = -1

    def _compute_from_obs_fcst(self, obs, fcst):
        sortedobs = np.sort(obs)
        sortedfcst = np.sort(fcst)
        return np.mean(np.abs(sortedobs - sortedfcst))


class FCrps(Metric):
    """ Fair continuous ranked probability score """
    type = verif.metric_type.Probabilistic()
    name = "Fair continuous ranked probability score"
    description = "Fair continuous ranked probability score"
    supports_aggregator = True
    orientation = -1

    def compute_single(self, data, input_index, axis, axis_index, interval):
        ensemble, obs = data.get_scores([verif.field.Ensemble(), verif.field.Obs()], input_index, axis, axis_index)
        if ensemble.shape[1] < 2:
            verif.util.error("Cannot compute CRPS with less than 2 members")
        crps = self.compute_crps(ensemble, obs)
        crps = self.aggregator(crps)
        return crps

    def label(self, variable):
        return "fCRPS (" + variable.units + ")"

    @staticmethod
    def compute_crps(ensemble, obs, fair=True):
        """Continuous Ranked Probability Score (CRPS).

        Args:
            ensemble: numpy.ndarray
                Predictions, shape (time, leadtime, location, ens_size)
            obs: numpy.ndarray
                Targets, shape (time, leadtime, location)
            fair: bool
                Defaults to true

        Returns:
            crps: numpy.ndarray
                Shape (time, leadtime, location)
        """
        num_members = ensemble.shape[-1]

        coef = (
            -1.0 / (num_members * (num_members - 1))
            if fair
            else -1.0 / (num_members**2)
        )

        mae = np.mean(np.abs(obs[..., None] - ensemble), axis=-1)

        # var = np.abs(ensemble[..., None] - ensemble[..., None, :])
        var = np.zeros(ensemble.shape[:-1])
        for i in range(num_members):  # loop version to reduce memory usage
            var += np.sum(np.abs(ensemble[..., i, None] - ensemble[..., i + 1:]), axis=-1)
        var *= coef
        return mae + var


class Pit(Metric):
    """ Retrives the PIT-value corresponding to the observation """
    type = verif.metric_type.Probabilistic()
    name = "Probability integral transform"
    description = "Verifying PIT-value (CDF at observation)"
    supports_aggregator = True
    orientation = 0

    def compute_single(self, data, input_index, axis, axis_index, interval):
        pit = data.get_scores(verif.field.Pit(), input_index, axis, axis_index)
        return self.aggregator(pit)

    def label(self, variable):
        return self.aggregator.name().title() + " of verifying PIT"


class PitHistDev(Metric):
    type = verif.metric_type.Probabilistic()
    name = "PIT histogram deviation factor"
    description = "PIT histogram deviation factor (actual deviation / expected deviation)"
    min = 0
    # max = 1
    perfect_score = 1
    orientation = -1

    def __init__(self, numBins=11, field=verif.field.Pit()):
        self._bins = np.linspace(0, 1, numBins)
        self._field = field

    def compute_single(self, data, input_index, axis, axis_index, interval):
        pit = data.get_scores(self._field, input_index, axis, axis_index)
        nb = len(self._bins) - 1
        D = self.deviation(pit, nb)
        D0 = self.expected_deviation(pit, nb)
        dev = D / D0
        return dev

    def label(self, variable):
        return self.name

    @staticmethod
    def expected_deviation(values, numBins):
        if len(values) == 0 or numBins == 0:
            return np.nan
        return np.sqrt((1.0 - 1.0 / numBins) / (len(values) * numBins))

    @staticmethod
    def deviation(values, numBins):
        if len(values) == 0 or numBins == 0:
            return np.nan
        x = np.linspace(0, 1, numBins + 1)
        n = np.histogram(values, x)[0]
        n = n * 1.0 / sum(n)
        return np.sqrt(1.0 / numBins * np.sum((n - 1.0 / numBins) ** 2))

    @staticmethod
    def deviation_std(values, numBins):
        if len(values) == 0 or numBins == 0:
            return np.nan
        n = len(values)
        p = 1.0 / numBins
        numPerBinStd = np.sqrt(n * p * (1 - p))
        std = numPerBinStd / n
        return std

    # What reduction in ignorance is possible by calibrating the PIT-histogram?
    @staticmethod
    def ignorance_potential(values, numBins):
        if len(values) == 0 or numBins == 0:
            return np.nan
        x = np.linspace(0, 1, numBins + 1)
        n = np.histogram(values, x)[0]
        n = n * 1.0 / sum(n)
        expected = 1.0 / numBins
        ign = np.sum(n * np.log2(n / expected)) / sum(n)
        return ign


class PitHistSlope(Metric):
    type = verif.metric_type.Probabilistic()
    name = "PIT histogram slope"
    description = "Average slope of the PIT histogram. Positive mean too many obs in the higher ranks."
    perfect_score = 0
    orientation = 0

    def __init__(self, numBins=11, field=verif.field.Pit()):
        self._bins = np.linspace(0, 1, numBins)
        self._field = field

    def compute_single(self, data, input_index, axis, axis_index, interval):
        # Create a PIT histogram, then compute the average slope across the bars
        pit = data.get_scores(self._field, input_index, axis, axis_index)
        n = np.histogram(pit, self._bins)[0]
        n = n * 1.0 / sum(n)

        centers = (self._bins[1:] + self._bins[0:-1]) / 2
        dx = np.diff(centers)
        d = np.diff(n) / dx
        return np.mean(d)

    def label(self, variable):
        return self.name


class PitHistShape(Metric):
    type = verif.metric_type.Probabilistic()
    name = "PIT histogram shape"
    description = "Second derivative of the PIT histogram. Negative means U-shaped."
    perfect_score = 0
    orientation = 0

    def __init__(self, numBins=11, field=verif.field.Pit()):
        self._bins = np.linspace(0, 1, numBins)
        self._field = field

    def compute_single(self, data, input_index, axis, axis_index, interval):
        # Create a PIT histogram, then compute the second derivative across the bars
        pit = data.get_scores(self._field, input_index, axis, axis_index)
        n = np.histogram(pit, self._bins)[0]
        n = n * 1.0 / sum(n)

        centers = (self._bins[1:] + self._bins[0:-1]) / 2
        dx = np.diff(centers)
        d = np.diff(n) / dx
        centers2 = (centers[1:] + centers[0:-1]) / 2
        dx2 = np.diff(centers2)
        dd = np.diff(d) / dx2
        return np.mean(dd)

    def label(self, variable):
        return self.name


class MarginalRatio(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Marginal ratio"
    description = "Ratio of marginal probability of obs to marginal" \
          " probability of fcst. Use -r to specify thresholds."
    min = 0
    perfect_score = 1
    require_threshold_type = "threshold"
    supports_threshold = True
    default_axis = verif.axis.Threshold()
    orientation = 0

    def compute_single(self, data, input_index, axis, axis_index, interval):
        if np.isinf(interval.lower):
            pvar = verif.field.Threshold(interval.upper)
            [obs, p1] = data.get_scores([verif.field.Obs(), pvar], input_index, axis, axis_index)
            p0 = 0 * p1
        elif np.isinf(interval.upper):
            pvar = verif.field.Threshold(interval.lower)
            [obs, p0] = data.get_scores([verif.field.Obs(), pvar], input_index,
                  axis, axis_index)
            p1 = 0 * p0 + 1
        else:
            pvar0 = verif.field.Threshold(interval.lower)
            pvar1 = verif.field.Threshold(interval.upper)
            [obs, p0, p1] = data.get_scores([verif.field.Obs(), pvar0, pvar1],
                  input_index, axis, axis_index)
        obs = interval.within(obs)
        p = p1 - p0
        if np.mean(p) == 0:
            return np.nan
        return np.mean(obs) / np.mean(p)

    def label(self, variable):
        return "Ratio of marginal probs: Pobs/Pfcst"


class Within(Metric):
    type = verif.metric_type.Deterministic()
    """ Can't be a subclass of ObsFcstBased, because it depends on threshold
    """
    name = "Within"
    description = "The percentage of forecasts within some error bound. Use -r to specify error bounds"
    min = 0
    max = 100
    default_bin_type = "below"
    require_threshold_type = "threshold"
    supports_threshold = True
    perfect_score = 100
    orientation = 0

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obs, fcst] = data.get_scores([verif.field.Obs(),
           verif.field.Fcst()], input_index, axis, axis_index)

        return self.compute_from_obs_fcst(obs, fcst, interval)

    def compute_from_obs_fcst(self, obs, fcst, interval):
        diff = abs(obs - fcst)
        return np.mean(interval.within(diff)) * 100

    def label(self, variable):
        return "% of forecasts"


class Conditional(Metric):
    """
    Computes the mean y conditioned on x. For a given range of x-values, what is
    the average y-value?
    """
    type = verif.metric_type.Deterministic()
    orientation = 0

    def __init__(self, x=verif.field.Obs(), y=verif.field.Fcst(), func=np.mean):
        self._x = x
        self._y = y
        self._func = func

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obs, fcst] = data.get_scores([self._x, self._y], input_index, axis, axis_index)
        return self.compute_from_obs_fcst(obs, fcst, interval)

    def compute_from_obs_fcst(self, obs, fcst, interval):
        I = np.where(interval.within(obs))[0]
        if len(I) == 0:
            return np.nan
        return self._func(fcst[I])


class XConditional(Metric):
    """
    Mean x when conditioned on x. Average x-value that is within a given range.
    The reason the y-variable is added is to ensure that the same data is used
    for this metric as for the Conditional metric.
    """
    type = verif.metric_type.Deterministic()
    orientation = 0

    def __init__(self, x=verif.field.Obs(), y=verif.field.Fcst(), func=np.median):
        self._x = x
        self._y = y
        self._func = func

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obs, fcst] = data.get_scores([self._x, self._y], input_index, axis, axis_index)
        return self.compute_from_obs_fcst(obs, fcst, interval)

    def compute_from_obs_fcst(self, obs, fcst, interval):
        I = np.where(interval.within(obs))[0]
        if len(I) == 0:
            return np.nan
        return self._func(obs[I])


class Count(Metric):
    """
    Counts how many values of a specific variable is within the threshold range
    Not a real metric.
    """
    type = verif.metric_type.Deterministic()
    orientation = 0

    def __init__(self, x):
        self._x = x

    def compute_single(self, data, input_index, axis, axis_index, interval):
        values = data.get_scores(self._x, input_index, axis, axis_index)
        I = np.where(interval.within(values))[0]
        if len(I) == 0:
            return np.nan
        return len(I)


class Bs(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Brier score"
    description = "Brier score"
    min = 0
    max = 1
    default_axis = verif.axis.Threshold()
    require_threshold_type = "threshold"
    supports_threshold = True
    perfect_score = 0
    orientation = -1
    reference = "Glenn W. Brier, 1950: Verification of forecasts expressed in terms of probability. Mon. Wea. Rev., 78, 1-3."

    def compute_single(self, data, input_index, axis, axis_index, interval):
        """ Compute probabilities based on thresholds """
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)
        return self.compute_from_obs_fcst(obsP, p)

    def compute_from_obs_fcst(self, obs, fcst):
        bs = np.nanmean((fcst-obs)**2)
        return bs

    def label(self, variable):
        return self.name


class BsRel(Metric):
    default_axis = verif.axis.Threshold()
    type = verif.metric_type.Probabilistic()
    name = "Brier score, reliability term"
    description = "Brier score, reliability term"
    min = 0
    max = 1
    require_threshold_type = "threshold"
    supports_threshold = True
    perfect_score = 0
    orientation = -1

    def __init__(self, num_edges=11):
        self._edges = np.linspace(0, 1, num_edges)
        self._edges[-1] = 1.001

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)
        return self.compute_from_obs_fcst(obsP, p)

    def compute_from_obs_fcst(self, obs, fcst):
        bs = np.nan * np.zeros(len(fcst), 'float')
        obs_mean = np.mean(obs)
        """
        Break p into bins, and compute reliability. but save each reliability
        value in an array the same size as fcst. In this way we do not need to do
        a weighted average
        """
        for i in range(0, len(self._edges) - 1):
            I = np.where((fcst >= self._edges[i]) & (fcst < self._edges[i + 1]))[0]
            if len(I) > 0:
                obs_mean_I = np.mean(obs[I])
                bs[I] = (fcst[I] - obs_mean_I) ** 2
        return np.nanmean(bs)

    def label(self, variable):
        return self.name


class BsRes(Metric):
    default_axis = verif.axis.Threshold()
    type = verif.metric_type.Probabilistic()
    name = "Brier score, resolution term"
    description = "Brier score, resolution term"
    min = 0
    max = 1
    require_threshold_type = "threshold"
    supports_threshold = True
    perfect_score = None  # Perfect score equals BsUnc
    orientation = 1

    def __init__(self, num_edges=11):
        self._edges = np.linspace(0, 1, num_edges)
        self._edges[-1] = 1.001

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)
        return self.compute_from_obs_fcst(obsP, p)

    def compute_from_obs_fcst(self, obs, fcst):
        bs = np.nan * np.zeros(len(fcst), 'float')
        obs_mean = np.mean(obs)
        for i in range(0, len(self._edges) - 1):
            I = np.where((fcst >= self._edges[i]) & (fcst < self._edges[i + 1]))[0]
            if len(I) > 0:
                obs_mean_I = np.mean(obs[I])
                bs[I] = (obs_mean_I - obs_mean) ** 2
        return np.nanmean(bs)

    def label(self, variable):
        return self.name


class BsUnc(Metric):
    default_axis = verif.axis.Threshold()
    type = verif.metric_type.Probabilistic()
    name = "Brier score, uncertainty term"
    description = "Brier score, uncertainty term"
    min = 0
    max = 1
    require_threshold_type = "threshold"
    supports_threshold = True
    perfect_score = None
    orientation = 0

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)
        return self.compute_from_obs_fcst(obsP, p)

    def compute_from_obs_fcst(self, obs, fcst):
        obs_mean = np.mean(obs)
        bsunc = np.nanmean((obs_mean - obs)**2)
        return bsunc

    def label(self, variable):
        return self.name


class Bss(Metric):
    default_axis = verif.axis.Threshold()
    type = verif.metric_type.Probabilistic()
    name = "Brier skill score"
    description = "Brier skill score"
    min = 0
    max = 1
    require_threshold_type = "threshold"
    supports_threshold = True
    perfect_score = 1
    orientation = 1

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)
        return self.compute_from_obs_fcst(obsP, p)

    def compute_from_obs_fcst(self, obs, fcst):
        bs = np.nanmean((fcst - obs)**2)
        obs_mean = np.mean(obs)
        bsunc = np.nanmean((obs_mean - obs)**2)
        if bsunc == 0:
            bss = np.nan
        else:
            bss = (bsunc - bs) / bsunc
        return bss

    def label(self, variable):
        return self.name


class BssRel(Metric):
    default_axis = verif.axis.Threshold()
    type = verif.metric_type.Probabilistic()
    name = "Brier skill score, reliability term"
    description = "Brier skill score, reliability term"
    min = 0
    max = 1
    require_threshold_type = "threshold"
    supports_threshold = True
    perfect_score = 0
    orientation = -1

    def __init__(self, num_edges=11):
        self._edges = np.linspace(0, 1, num_edges)
        self._edges[-1] = 1.001

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)
        return self.compute_from_obs_fcst(obsP, p)

    def compute_from_obs_fcst(self, obs, fcst):
        bsrel = np.nan * np.zeros(len(fcst), 'float')
        obs_mean = np.mean(obs)
        for i in range(0, len(self._edges) - 1):
            I = np.where((fcst >= self._edges[i]) & (fcst < self._edges[i + 1]))[0]
            if len(I) > 0:
                obs_mean_I = np.mean(obs[I])
                bsrel[I] = (fcst[I] - obs_mean_I) ** 2
        bsunc = np.nanmean((obs_mean - obs)**2)
        bsrel = np.nanmean(bsrel)
        if bsunc == 0:
            bss = np.nan
        else:
            bss = bsrel / bsunc
        return bss

    def label(self, variable):
        return self.name


class BssRes(Metric):
    default_axis = verif.axis.Threshold()
    type = verif.metric_type.Probabilistic()
    name = "Brier skill score, resolution term"
    description = "Brier skill score, resolution term"
    min = 0
    max = 1
    require_threshold_type = "threshold"
    supports_threshold = True
    perfect_score = 1
    orientation = 1

    def __init__(self, num_edges=11):
        self._edges = np.linspace(0, 1, num_edges)
        self._edges[-1] = 1.001

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)
        return self.compute_from_obs_fcst(obsP, p)

    def compute_from_obs_fcst(self, obs, fcst):
        bsres = np.nan * np.zeros(len(fcst), 'float')
        obs_mean = np.mean(obs)
        for i in range(0, len(self._edges) - 1):
            I = np.where((fcst >= self._edges[i]) & (fcst < self._edges[i + 1]))[0]
            if len(I) > 0:
                obs_mean_I = np.mean(obs[I])
                bsres[I] = (obs_mean_I - obs_mean) ** 2
        bsres = np.nanmean(bsres)
        bsunc = np.nanmean((obs_mean - obs)**2)
        if bsunc == 0:
            bss = np.nan
        else:
            bss = bsres / bsunc
        return bss

    def label(self, variable):
        return self.name


class QuantileCoverage(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Capture"
    description = "Fraction of observations contained within quantile interval. Use -q to set quantiles, and -b to set interval definition."
    min = 0
    # default_bin_type = "within"
    require_threshold_type = "quantile"
    supports_threshold = True
    min_num_thresholds = 1
    max_num_thresholds = 2

    def compute_single(self, data, input_index, axis, axis_index, interval):
        var0 = verif.field.Quantile(interval.lower)
        var1 = verif.field.Quantile(interval.upper)
        if np.isinf(interval.lower):
            [obs, q1] = data.get_scores([verif.field.Obs(), var1], input_index, axis, axis_index)
            I = np.where((np.isnan(obs) == 0) & (np.isnan(q1) == 0))[0]
            if interval.upper_eq:
                return np.mean(q1[I] >= obs[I])
            else:
                return np.mean(q1[I] > obs[I])
        elif np.isinf(interval.upper):
            [obs, q0] = data.get_scores([verif.field.Obs(), var0], input_index, axis, axis_index)
            I = np.where((np.isnan(obs) == 0) & (np.isnan(q0) == 0))[0]
            if interval.lower_eq:
                return np.mean(q0[I] <= obs[I])
            else:
                return np.mean(q0[I] < obs[I])
        else:
            [obs, q0, q1] = data.get_scores([verif.field.Obs(), var0, var1], input_index, axis, axis_index)
            I = np.where((np.isnan(obs) == 0) & (np.isnan(q0) == 0) & (np.isnan(q1) == 0))[0]
            if interval.lower_eq:
                c0 = q0[I] <= obs[I]
            else:
                c0 = q0[I] < obs[I]
            if interval.upper_eq:
                c1 = q1[I] >= obs[I]
            else:
                c1 = q1[I] > obs[I]
            return np.mean(c0 & c1)

    def label(self, variable):
        return self.name


class Quantile(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Quantile"
    description = "Mean value of a quantile forecast. Use -q to set quantile."
    require_threshold_type = "quantile"
    supports_threshold = True

    def compute_single(self, data, input_index, axis, axis_index, interval):
        if not np.isinf(interval.lower) and not np.isinf(interval.upper):
            var0 = verif.field.Quantile(interval.lower)
            var1 = verif.field.Quantile(interval.upper)
            t0, t1 = data.get_scores([var0, var1], input_index, axis, axis_index)
            I = np.where((np.isnan(t0) == 0) & (np.isnan(t1) == 0))[0]
            return np.mean(t1[I] - t0[I])
        elif np.isinf(interval.lower):
            var = verif.field.Quantile(interval.upper)
        elif np.isinf(interval.upper):
            var = verif.field.Quantile(interval.lower)

        t = data.get_scores(var, input_index, axis, axis_index)
        I = np.where(np.isnan(t) == 0)[0]
        return np.mean(t[I])

    def label(self, variable):
        return self.aggregator.name().title() + " of quantile forecast (" + variable.units + ")"


class Threshold(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Threshold"
    description = "Mean value of threshold forecast. Use -r to set threshold."
    min = 0
    require_threshold_type = "thresholds"
    supports_threshold = True

    def compute_single(self, data, input_index, axis, axis_index, interval):
        var0 = verif.field.Threshold(interval.lower)
        var1 = verif.field.Threshold(interval.upper)
        if not np.isinf(interval.lower) and not np.isinf(interval.upper):
            t0, t1 = data.get_scores([var0, var1], input_index, axis, axis_index)
            I = np.where((np.isnan(t0) == 0) & (np.isnan(t1) == 0))[0]
            return np.mean(t1[I] - t0[I])
        elif np.isinf(interval.lower):
            use_var = var1
        elif np.isinf(interval.upper):
            use_var = var0

        t = data.get_scores(use_var, input_index, axis, axis_index)
        I = np.where(np.isnan(t) == 0)[0]
        return np.mean(t[I])

    def label(self, variable):
        return self.aggregator.name().title() + " of threshold probability"


class QuantileScore(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Quantile score"
    description = "Quantile score. Use -q to set which quantiles to use, and -b to set interval definition."
    min = 0
    require_threshold_type = "quantile"
    supports_threshold = True
    perfect_score = 0
    orientation = -1

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obs, pred_q] = get_q(data, input_index, axis, axis_index, interval)
        err = obs - pred_q
        qs = err * (interval.lower - (err < 0))
        return np.mean(qs)

    def label(self, variable):
        return self.name


class Spread(FromField):
    type = verif.metric_type.Probabilistic()
    name = "Ensemble standard deviation"
    description = "Ensemble spread (standard deviation of members)."
    min = 0
    supports_aggregator = True
    perfect_score = 0
    orientation = -1

    def __init__(self):
        super(Spread, self).__init__(verif.field.EnsembleVariance())

    def label(self, variable):
        return self.aggregator.name().title() + " of ensemble spread (" + variable.units + ")"


class SpreadSkillRatio(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Spread-skill ratio"
    description = "Ratio of spread (standard deviation of memebrs) to skill (RMSE of ensemble mean)."
    supports_threshold = False
    min = 0
    perfect_score = 1
    orientation = 0

    def compute_single(self, data, input_index, axis, axis_index, interval):
        fields = [verif.field.EnsembleMean(), verif.field.EnsembleVariance(), verif.field.Obs()]
        mean, variance, obs = data.get_scores(fields, input_index, axis, axis_index)

        spread = np.mean(np.sqrt(variance))
        skill = np.sqrt(np.mean(np.abs(obs - mean)**2))

        if np.isclose(skill, 0, atol=1e-7):
            return np.nan
        else:
            return spread / skill

    def label(self, variable):
        return self.name


class Ign0(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Binary ignorance"
    description = "Ignorance of the binary probability based on threshold"
    require_threshold_type = "threshold"
    supports_threshold = True
    orientation = -1

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)

        I0 = np.where(obsP == 0)[0]
        I1 = np.where(obsP == 1)[0]
        ign = -np.log2(p)
        ign[I0] = -np.log2(1 - p[I0])
        return np.mean(ign)

    def label(self, variable):
        return self.name


class Spherical(Metric):
    type = verif.metric_type.Probabilistic()
    name = "Spherical score"
    description = "Spherical probabilistic scoring rule for binary events"
    require_threshold_type = "threshold"
    supports_threshold = True
    max = 1
    min = 0
    perfect_score = 1
    orientation = 1

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obsP, p] = get_p(data, input_index, axis, axis_index, interval)

        I0 = np.where(obsP == 0)[0]
        I1 = np.where(obsP == 1)[0]
        sp = p / np.sqrt(p ** 2 + (1 - p) ** 2)
        sp[I0] = (1 - p[I0]) / np.sqrt((p[I0]) ** 2 + (1 - p[I0]) ** 2)

        return np.mean(sp)

    def label(self, variable):
        return self.name


class Contingency(Metric):
    """
    Metrics based on 2x2 contingency table for a given interval. Observations
    and forecasts are converted into binary values, that is if they are within
    or not within an interval.
    """
    type = verif.metric_type.Threshold()
    min = 0
    max = 1
    default_axis = verif.axis.Threshold()
    require_threshold_type = "deterministic"
    supports_threshold = True
    _usingQuantiles = False

    def compute_from_abcd(self, a, b, c, d):
        """ Compute the score given the 4 values in the 2x2 contingency table:

        Arguments:
           a (float): Hit
           b (float): False alarm
           c (float): Miss
           d (float): Correct rejection

        Returns:
           float: The score
        """
        raise NotImplementedError()

    def label(self, variable):
        return self.name

    def compute_single(self, data, input_index, axis, axis_index, interval):
        [obs, fcst] = data.get_scores([verif.field.Obs(), verif.field.Fcst()], input_index, axis, axis_index)
        return self.compute_from_obs_fcst(obs, fcst, interval)

    def _quantile_to_threshold(self, values, interval):
        """
        Convert an interval of quantiles to interval thresholds, for example
        converting [10%, 50%] of some precip values to [5 mm, 25 mm]

        Arguments:
           values (np.array): values to compute thresholds for
           interval (verif.interval.Interval): interval of quantiles

        Returns:
           verif.interval.Interval: Interval of thresholds
        """
        sorted = np.sort(values)
        lower = -np.inf
        upper = np.inf
        if not np.isinf(abs(interval.lower)):
            lower = np.percentile(sorted, interval.lower * 100)
        if not np.isinf(abs(interval.lower)):
            upper = np.percentile(sorted, interval.upper * 100)
        return verif.interval.Interval(lower, upper, interval.lower_eq, interval.upper_eq)

    def _compute_abcd(self, obs, fcst, interval, f_interval=None):
        if f_interval is None:
            f_interval = interval
        a = b = c = d = np.nan
        if len(fcst) > 0:
            # Compute frequencies
            if self._usingQuantiles:
                fcstSort = np.sort(fcst)
                obsSort = np.sort(obs)
                f_qinterval = self._quantile_to_threshold(fcstSort, f_interval)
                o_qinterval = self._quantile_to_threshold(obsSort, interval)
                a = np.ma.sum(f_qinterval.within(fcst) & o_qinterval.within(obs))  # Hit
                b = np.ma.sum(f_qinterval.within(fcst) & (o_qinterval.within(obs) == 0))  # FA
                c = np.ma.sum((f_qinterval.within(fcst) == 0) & o_qinterval.within(obs))  # Miss
                d = np.ma.sum((f_qinterval.within(fcst) == 0) & (o_qinterval.within(obs) == 0))  # CR
            else:
                a = np.ma.sum(f_interval.within(fcst) & interval.within(obs))  # Hit
                b = np.ma.sum(f_interval.within(fcst) & (interval.within(obs) == 0))  # FA
                c = np.ma.sum((f_interval.within(fcst) == 0) & interval.within(obs))  # Miss
                d = np.ma.sum((f_interval.within(fcst) == 0) & (interval.within(obs) == 0))  # CR
        return [a, b, c, d]

    def compute_from_obs_fcst(self, obs, fcst, interval, f_interval=None):
        """ Computes the score

        Arguments:
           obs (np.array): array of observations
           fcst (np.array): array of forecasts
           interval (verif.interval.Interval): compute score for this interval
           f_interval (verif.interval.Interval): Use this interval for forecasts.
              If None, then use the same interval for obs and forecasts.

        Returns:
           float: The score
        """
        [a, b, c, d] = self._compute_abcd(obs, fcst, interval, f_interval)

        value = self.compute_from_abcd(a, b, c, d)
        if np.isinf(value):
            value = np.nan

        return value

    def compute_from_obs_fcst_resample(self, obs, fcst, N, interval, f_interval=None):
        """
        Same as compute_from_obs_fcst, except compute more robust scores by
        resampling (with replacement) using the computed values of a, b, c, d.

        Arguments:
           obs (np.array): array of observations
           fcst (np.array): array of forecasts
           N (int): Resample this many times
           interval (verif.interval.Interval): compute score for this interval
           f_interval (verif.interval.Interval): Use this interval for forecasts.
              If None, then use the same interval for obs and forecasts.

        Returns:
           float: The score
        """
        [a, b, c, d] = self._compute_abcd(obs, fcst, interval, f_interval)

        # Resample
        n = a + b + c + d
        np.random.seed(1)
        value = 0
        for i in range(0, N):
            aa = np.random.binomial(n, 1.0*a/n)
            bb = np.random.binomial(n, 1.0*b/n)
            cc = np.random.binomial(n, 1.0*c/n)
            dd = np.random.binomial(n, 1.0*d/n)
            value = value + self.compute_from_abcd(aa, bb, cc, dd)
        value = value / N
        return value

    def label(self, variable):
        return self.name


class A(Contingency):
    name = "Hit"
    description = "Hit"

    def compute_from_abcd(self, a, b, c, d):
        return 1.0 * a / (a + b + c + d)


class B(Contingency):
    name = "False alarm"
    description = "False alarm"

    def compute_from_abcd(self, a, b, c, d):
        return 1.0 * b / (a + b + c + d)


class C(Contingency):
    name = "Miss"
    description = "Miss"

    def compute_from_abcd(self, a, b, c, d):
        return 1.0 * c / (a + b + c + d)


class D(Contingency):
    name = "Correct rejection"
    description = "Correct rejection"

    def compute_from_abcd(self, a, b, c, d):
        return 1.0 * d / (a + b + c + d)


class N(Contingency):
    name = "Total cases"
    description = "Total cases"
    max = None

    def compute_from_abcd(self, a, b, c, d):
        return a + b + c + d


class Ets(Contingency):
    name = "Equitable threat score"
    description = "Equitable threat score"
    perfect_score = 1
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        N = a + b + c + d
        ar = (a + b) / 1.0 / N * (a + c)
        if a + b + c - ar == 0:
            return np.nan
        return (a - ar) / 1.0 / (a + b + c - ar)

    def label(self, variable):
        return "ETS"


class FcstRate(Contingency):
    name = "Forecast rate"
    description = "Fractions of forecasts (a + b)"
    perfect_score = None
    orientation = 0

    def compute_from_abcd(self, a, b, c, d):
        return (a + b) / 1.0 / (a + b + c + d)


class Dscore(Contingency):
    name = "Discimination"
    description = "Generalized discrimination score"
    perfect_score = 1
    orientation = 1
    reference = "Simon J. Mason and Andreas P. Weigel, 2009: A Generic Forecast Verification Framework for Administrative Purposes. Mon. Wea. Rev., 137, 331-349."
    max = 1
    min = 0

    def compute_from_abcd(self, a, b, c, d):
        N = a + b + c + d
        num = a*d + 0.5*(a*b + c*d)
        denom = (a + c) * (b + d)
        if denom == 0:
            return np.nan
        return num / denom


class Threat(Contingency):
    name = "Threat score"
    description = "Threat score"
    perfect_score = 1
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        if a + b + c == 0:
            return np.nan
        return a / 1.0 / (a + b + c)


class Pc(Contingency):
    name = "Proportion correct"
    description = "Proportion correct"
    perfect_score = 1
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        return (a + d) / 1.0 / (a + b + c + d)


class Edi(Contingency):
    name = "Extremal dependency index"
    description = "Extremal dependency index"
    perfect_score = 1
    orientation = 1
    reference = "Christopher A. T. Ferro and David B. Stephenson, 2011: Extremal Dependence Indices: Improved Verification Measures for Deterministic Forecasts of Rare Binary Events. Wea. Forecasting, 26, 699-713."

    def compute_from_abcd(self, a, b, c, d):
        N = a + b + c + d
        if b + d == 0 or a + c == 0:
            return np.nan
        F = b / 1.0 / (b + d)
        H = a / 1.0 / (a + c)
        if H == 0 or F == 0:
            return np.nan
        denom = (np.log(H) + np.log(F))
        if denom == 0:
            return np.nan
        return (np.log(F) - np.log(H)) / denom

    def label(self, variable):
        return "EDI"


class Sedi(Contingency):
    name = "Symmetric extremal dependency index"
    description = "Symmetric extremal dependency index"
    perfect_score = 1
    orientation = 1
    reference = Edi.reference

    def compute_from_abcd(self, a, b, c, d):
        N = a + b + c + d
        if b + d == 0 or a + c == 0:
            return np.nan
        F = b / 1.0 / (b + d)
        H = a / 1.0 / (a + c)
        if F == 0 or F == 1 or H == 0 or H == 1:
            return np.nan
        denom = np.log(F) + np.log(H) + np.log(1 - F) + np.log(1 - H)
        if denom == 0:
            return np.nan
        num = np.log(F) - np.log(H) - np.log(1 - F) + np.log(1 - H)
        return num / denom

    def label(self, variable):
        return "SEDI"


class Eds(Contingency):
    name = "Extreme dependency score"
    description = "Extreme dependency score"
    min = None
    perfect_score = 1
    orientation = 1
    reference = "Stephenson, D. B., B. Casati, C. A. T. Ferro, and C. A.  Wilson, 2008: The extreme dependency score: A non-vanishing measure for forecasts of rare events. Meteor. Appl., 15, 41-50."

    def compute_from_abcd(self, a, b, c, d):
        N = a + b + c + d
        if a + c == 0:
            return np.nan
        H = a / 1.0 / (a + c)
        p = (a + c) / 1.0 / N
        if H == 0 or p == 0:
            return np.nan
        denom = (np.log(p) + np.log(H))
        if denom == 0:
            return np.nan
        return (np.log(p) - np.log(H)) / denom

    def label(self, variable):
        return "EDS"


class Seds(Contingency):
    name = "Symmetric extreme dependency score"
    description = "Symmetric extreme dependency score"
    min = None
    perfect_score = 1
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        N = a + b + c + d
        if a + c == 0:
            return np.nan
        H = a / 1.0 / (a + c)
        p = (a + c) / 1.0 / N
        q = (a + b) / 1.0 / N
        if q == 0 or H == 0:
            return np.nan
        denom = np.log(p) + np.log(H)
        if denom == 0:
            return np.nan
        return (np.log(q) - np.log(H)) / (np.log(p) + np.log(H))

    def label(self, variable):
        return "SEDS"


class BiasFreq(Contingency):
    name = "Bias frequency"
    description = "Bias frequency (number of fcsts / number of obs)"
    max = None
    perfect_score = 1
    orientation = 0

    def compute_from_abcd(self, a, b, c, d):
        if a + c == 0:
            return np.nan
        return 1.0 * (a + b) / (a + c)


class Hss(Contingency):
    name = "Heidke skill score"
    description = "Heidke skill score"
    max = 1
    perfect_score = 1
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        denom = ((a + c) * (c + d) + (a + b) * (b + d))
        if denom == 0:
            return np.nan
        return 2.0 * (a * d - b * c) / denom


class BaseRate(Contingency):
    name = "Base rate"
    description = "Base rate: Fraction of observations (a + c)"
    perfect_score = None
    orientation = 0

    def compute_from_abcd(self, a, b, c, d):
        if a + b + c + d == 0:
            return np.nan
        return (a + c) / 1.0 / (a + b + c + d)


class Or(Contingency):
    name = "Odds ratio"
    description = "Odds ratio"
    max = None
    perfect_score = None  # Should be infinity
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        if b * c == 0:
            return np.nan
        return (a * d) / 1.0 / (b * c)


class Lor(Contingency):
    name = "Log odds ratio"
    description = "Log odds ratio"
    max = None
    perfect_score = None  # Should be infinity
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        if a * d == 0 or b * c == 0:
            return np.nan
        return np.log((a * d) / 1.0 / (b * c))


class YulesQ(Contingency):
    name = "Yule's Q"
    description = "Yule's Q (Odds ratio skill score)"
    perfect_score = 1
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        if a * d + b * c == 0:
            return np.nan
        return (a * d - b * c) / 1.0 / (a * d + b * c)


class Kss(Contingency):
    name = "Hanssen-Kuiper skill score"
    description = "Hanssen-Kuiper skill score"
    perfect_score = 1
    orientation = 1
    reference = "Hanssen , A., W. Kuipers, 1965: On the relationship between the frequency of rain and various meteorological parameters. - Meded. Verh. 81, 2-15."

    def compute_from_abcd(self, a, b, c, d):
        if (a + c) * (b + d) == 0:
            return np.nan
        return (a * d - b * c) * 1.0 / ((a + c) * (b + d))


class Hit(Contingency):
    name = "Hit rate"
    description = "Hit rate (a.k.a. probability of detection)"
    perfect_score = 1
    orientation = 1

    def compute_from_abcd(self, a, b, c, d):
        if a + c == 0:
            return np.nan
        return a / 1.0 / (a + c)


class Miss(Contingency):
    name = "Miss rate"
    description = "Miss rate"
    perfect_score = 0
    orientation = -1

    def compute_from_abcd(self, a, b, c, d):
        if a + c == 0:
            return np.nan
        return c / 1.0 / (a + c)


# Fraction of non-events that are forecasted as events
class Fa(Contingency):
    name = "False alarm rate"
    description = "False alarm rate"
    perfect_score = 0
    orientation = -1

    def compute_from_abcd(self, a, b, c, d):
        if b + d == 0:
            return np.nan
        return b / 1.0 / (b + d)


# Fraction of forecasted events that are false alarms
class Far(Contingency):
    name = "False alarm ratio"
    description = "False alarm ratio"
    perfect_score = 0
    orientation = -1

    def compute_from_abcd(self, a, b, c, d):
        if a + b == 0:
            return np.nan
        return b / 1.0 / (a + b)



import os
import sys
import numpy as np
import matplotlib.pylab as mpl
import argparse
import verif.util
import netCDF4
import shutil
import scipy.interpolate


def main():
    parser = argparse.ArgumentParser(description='This program adds pre-computed ensemble statistics to a Verif file. It also renames variables using an older convention, if needed.')
    parser.add_argument('file', help='Input file')

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    # Make a copy, in case the append corrupts the file (shouldn't happen)
    filename_backup = args.file + ".tmp"
    shutil.copy(args.file, filename_backup)

    with netCDF4.Dataset(args.file, 'a') as file:
        # Rename variables
        if "crps" in file.variables:
            file.renameVariable("crps", "ensemble_crps")
        if "ens-mean" in file.variables:
            file.renameVariable("ens-mean", "ensemble_mean")
        if "ens-std" in file.variables:
            file.renameVariable("ens-std", "ensemble_std")
        if "ens-var" in file.variables:
            file.renameVariable("ens-var", "ensemble_variance")

        if "ensemble" not in file.variables:
            print(f"File does not have ensemble, cannot compute ensemble statistics.")
        else:
            ensemble = file.variables["ensemble"][:]
            num_members = ensemble.shape[-1]

            # Compute CRPS
            if "ensemble_crps" not in file.variables:
                file.createvariable("ensemble_crps", "f4", ("time", "leadtime", "location"))

            crps = compute_crps(ensemble, file.variables["obs"][:], num_members)
            num_valid_members = np.sum(~np.isnan(ensemble), axis=3)
            crps[num_valid_members != num_members] = np.nan
            file.variables["ensemble_crps"][:] = crps

            # Compute ensemble moments
            fields = ["ensemble_mean" , "ensemble_variance"]
            for field in fields:
                if field not in file.variables:
                    file.createVariable(field, "f4", ("time", "leadtime", "location"))

            ens_mean = np.nanmean(ensemble, axis=3)
            file.variables["ensemble_mean"][:] = ens_mean
            file.variables["ensemble_variance"][:] = np.nanvar(ensemble, axis=3)

    os.remove(filename_backup)


def compute_crps(preds, targets, num_members, fair=True) -> np.ndarray:
    """Continuous Ranked Probability Score (CRPS).

    Taken from bris-inference

    Args:
        preds: numpy.ndarray
            Predictions, shape (time, leadtime, location, ens_size)
        targets: numpy.ndarray
            Targets, shape (time, leadtime, location)
        fair: bool
            Defaults to true

    Returns:
        crps: numpy.ndarray
            Shape (time, leadtime, location)
    """

    coef = (
        -1.0 / (num_members * (num_members - 1))
        if fair
        else -1.0 / (num_members**2)
    )

    mae = np.mean(np.abs(targets[..., None] - preds), axis=-1)

    # var = np.abs(preds[..., None] - preds[..., None, :])
    var = np.zeros(preds.shape[:-1])
    for i in range(num_members):  # loop version to reduce memory usage
        var += np.sum(np.abs(preds[..., i, None] - preds[..., i + 1 :]), axis=-1)
    var *= coef
    return mae + var


if __name__ == "__main__":
    main()
